# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/feature_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import basic_info__client_pb2 as tecton__proto_dot_args_dot_basic__info____client__pb2
from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2
from tecton_proto.common import framework_version__client_pb2 as tecton__proto_dot_common_dot_framework__version____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/tecton_proto/args/feature_service__client.proto\x12\x11tecton_proto.args\x1a*tecton_proto/args/basic_info__client.proto\x1a,tecton_proto/args/diff_options__client.proto\x1a\x33tecton_proto/common/framework_version__client.proto\x1a$tecton_proto/common/id__client.proto\"d\n\x11LoggingConfigArgs\x12\x1f\n\x0bsample_rate\x18\x01 \x01(\x02R\nsampleRate\x12.\n\x13log_effective_times\x18\x02 \x01(\x08R\x11logEffectiveTimes\"V\n\nColumnPair\x12!\n\x0cspine_column\x18\x01 \x01(\tR\x0bspineColumn\x12%\n\x0e\x66\x65\x61ture_column\x18\x02 \x01(\tR\rfeatureColumn\"\xe1\x01\n\x10\x46\x65\x61tureReference\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12R\n\x12override_join_keys\x18\x02 \x03(\x0b\x32\x1d.tecton_proto.args.ColumnPairB\x05\x92M\x02 \x07R\x10overrideJoinKeys\x12\x1c\n\tnamespace\x18\x03 \x01(\tR\tnamespace\x12\x1a\n\x08\x66\x65\x61tures\x18\x04 \x03(\tR\x08\x66\x65\x61tures\"\xe2\x05\n\x12\x46\x65\x61tureServiceArgs\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12\x37\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12\x46\n\x07version\x18\x06 \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionB\x05\x92M\x02\x08\x05R\x07version\x12.\n\x0fprevent_destroy\x18\x07 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x0epreventDestroy\x12S\n\x07options\x18\n \x03(\x0b\x32\x32.tecton_proto.args.FeatureServiceArgs.OptionsEntryB\x05\x92M\x02\x08\x01R\x07options\x12\x39\n\x15\x65nable_online_caching\x18\x0b \x01(\x08\x42\x05\x92M\x02\x08\x01R\x13\x65nableOnlineCaching\x12R\n\x12\x66\x65\x61ture_references\x18\x03 \x03(\x0b\x32#.tecton_proto.args.FeatureReferenceR\x11\x66\x65\x61tureReferences\x12;\n\x16online_serving_enabled\x18\x04 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x14onlineServingEnabled\x12>\n\x07logging\x18\x05 \x01(\x0b\x32$.tecton_proto.args.LoggingConfigArgsR\x07logging\x12\x31\n\x14realtime_environment\x18\t \x01(\tR\x13realtimeEnvironment\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x08\x10\tB\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.feature_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _FEATUREREFERENCE.fields_by_name['override_join_keys']._options = None
  _FEATUREREFERENCE.fields_by_name['override_join_keys']._serialized_options = b'\222M\002 \007'
  _FEATURESERVICEARGS_OPTIONSENTRY._options = None
  _FEATURESERVICEARGS_OPTIONSENTRY._serialized_options = b'8\001'
  _FEATURESERVICEARGS.fields_by_name['info']._options = None
  _FEATURESERVICEARGS.fields_by_name['info']._serialized_options = b'\222M\002\020\001'
  _FEATURESERVICEARGS.fields_by_name['version']._options = None
  _FEATURESERVICEARGS.fields_by_name['version']._serialized_options = b'\222M\002\010\005'
  _FEATURESERVICEARGS.fields_by_name['prevent_destroy']._options = None
  _FEATURESERVICEARGS.fields_by_name['prevent_destroy']._serialized_options = b'\222M\002\010\001'
  _FEATURESERVICEARGS.fields_by_name['options']._options = None
  _FEATURESERVICEARGS.fields_by_name['options']._serialized_options = b'\222M\002\010\001'
  _FEATURESERVICEARGS.fields_by_name['enable_online_caching']._options = None
  _FEATURESERVICEARGS.fields_by_name['enable_online_caching']._serialized_options = b'\222M\002\010\001'
  _FEATURESERVICEARGS.fields_by_name['online_serving_enabled']._options = None
  _FEATURESERVICEARGS.fields_by_name['online_serving_enabled']._serialized_options = b'\222M\002\010\001'
  _LOGGINGCONFIGARGS._serialized_start=251
  _LOGGINGCONFIGARGS._serialized_end=351
  _COLUMNPAIR._serialized_start=353
  _COLUMNPAIR._serialized_end=439
  _FEATUREREFERENCE._serialized_start=442
  _FEATUREREFERENCE._serialized_end=667
  _FEATURESERVICEARGS._serialized_start=670
  _FEATURESERVICEARGS._serialized_end=1408
  _FEATURESERVICEARGS_OPTIONSENTRY._serialized_start=1344
  _FEATURESERVICEARGS_OPTIONSENTRY._serialized_end=1402
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/fv_materialization__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import fco_locator__client_pb2 as tecton__proto_dot_common_dot_fco__locator____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/data/fv_materialization__client.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a-tecton_proto/common/fco_locator__client.proto\"\"\n\x0c\x46ileLocation\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\"\xa5\x08\n\x11\x46vMaterialization\x12X\n\x17id_feature_view_locator\x18\t \x01(\x0b\x32!.tecton_proto.common.IdFcoLocatorR\x14idFeatureViewLocator\x12\x44\n\x1ematerialization_serial_version\x18\x03 \x01(\x05R\x1cmaterializationSerialVersion\x12]\n\x1amaterialized_data_location\x18\x05 \x01(\x0b\x32\x1f.tecton_proto.data.FileLocationR\x18materializedDataLocation\x12\x65\n\x1estreaming_checkpoint_locations\x18\x06 \x03(\x0b\x32\x1f.tecton_proto.data.FileLocationR\x1cstreamingCheckpointLocations\x12\x32\n\x15has_materialized_data\x18\x07 \x01(\x08R\x13hasMaterializedData\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x08 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersion\x12\\\n\x1amisc_offline_data_location\x18\x0b \x01(\x0b\x32\x1f.tecton_proto.data.FileLocationR\x17miscOfflineDataLocation\x12V\n(online_table_imported_by_materialization\x18\x0c \x01(\x08R$onlineTableImportedByMaterialization\x12?\n\x1conline_table_import_complete\x18\r \x01(\x08R\x19onlineTableImportComplete\x12`\n\x1c\x66\x65\x61ture_export_data_location\x18\x0e \x01(\x0b\x32\x1f.tecton_proto.data.FileLocationR\x19\x66\x65\x61tureExportDataLocation\x12P\n\x12online_data_tables\x18\x0f \x03(\x0b\x32\".tecton_proto.data.OnlineDataTableR\x10onlineDataTables\x12*\n\x11has_been_revisted\x18\x10 \x01(\x08R\x0fhasBeenRevisted\x12\x46\n\x11last_revisited_ts\x18\x11 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0flastRevisitedTsJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x04\x10\x05J\x04\x08\n\x10\x0b\"\xbb\x02\n\x0fOnlineDataTable\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12P\n\x16\x66\x65\x61ture_data_watermark\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x14\x66\x65\x61tureDataWatermark\x12^\n\x11state_transitions\x18\x03 \x03(\x0b\x32\x31.tecton_proto.data.OnlineDataTableStateTransitionR\x10stateTransitions\x12\x62\n\x1a\x62\x61tch_table_format_version\x18\x04 \x01(\x0e\x32%.tecton_proto.data.TableFormatVersionR\x17\x62\x61tchTableFormatVersion\"\x99\x01\n\x1eOnlineDataTableStateTransition\x12=\n\x05state\x18\x01 \x01(\x0e\x32\'.tecton_proto.data.OnlineDataTableStateR\x05state\x12\x38\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp*\xb5\x01\n\x16OnlineBackfillLoadType\x12)\n%ONLINE_BACKFILL_LOAD_TYPE_UNSPECIFIED\x10\x00\x12\"\n\x1eONLINE_BACKFILL_LOAD_TYPE_TASK\x10\x01\x12\"\n\x1eONLINE_BACKFILL_LOAD_TYPE_BULK\x10\x02\x12(\n$ONLINE_BACKFILL_LOAD_TYPE_COMPACTION\x10\x03*|\n\x12TableFormatVersion\x12,\n(TABLE_FORMAT_VERSION_DEFAULT_UNSPECIFIED\x10\x00\x12\x1b\n\x17TABLE_FORMAT_VERSION_V2\x10\x01\x12\x1b\n\x17TABLE_FORMAT_VERSION_V3\x10\x02*\xc1\x01\n\x14OnlineDataTableState\x12(\n$DATA_TABLE_STATE_UNKNOWN_UNSPECIFIED\x10\x00\x12\x1e\n\x1a\x44\x41TA_TABLE_STATE_NOT_READY\x10\x01\x12\x1a\n\x16\x44\x41TA_TABLE_STATE_READY\x10\x02\x12%\n!DATA_TABLE_STATE_PENDING_DELETION\x10\x03\x12\x1c\n\x18\x44\x41TA_TABLE_STATE_DELETED\x10\x04\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.fv_materialization__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _ONLINEBACKFILLLOADTYPE._serialized_start=1728
  _ONLINEBACKFILLLOADTYPE._serialized_end=1909
  _TABLEFORMATVERSION._serialized_start=1911
  _TABLEFORMATVERSION._serialized_end=2035
  _ONLINEDATATABLESTATE._serialized_start=2038
  _ONLINEDATATABLESTATE._serialized_end=2231
  _FILELOCATION._serialized_start=153
  _FILELOCATION._serialized_end=187
  _FVMATERIALIZATION._serialized_start=190
  _FVMATERIALIZATION._serialized_end=1251
  _ONLINEDATATABLE._serialized_start=1254
  _ONLINEDATATABLE._serialized_end=1569
  _ONLINEDATATABLESTATETRANSITION._serialized_start=1572
  _ONLINEDATATABLESTATETRANSITION._serialized_end=1725
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import attrs
import pendulum
import pyspark
from pyspark.sql.column import Column

from tecton_core import query_consts
from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.compaction_utils import AggregationGroup
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query_consts import udf_internal
from tecton_core.schema import Schema
from tecton_spark import data_observability
from tecton_spark import partial_aggregations
from tecton_spark.feature_view_spark_utils import validate_df_columns_and_feature_types
from tecton_spark.query.node import SparkExecNode
from tecton_spark.rtfv_pipeline import RealtimeFeaturePipeline
from tecton_spark.schema_spark_utils import schema_to_spark
from tecton_spark.spark_pipeline import SparkFeaturePipeline


@attrs.frozen
class MultiOdfvPipelineSparkNode(SparkExecNode):
    input_node: SparkExecNode
    use_namespace_feature_prefix: bool
    events_df_timestamp_field: str
    feature_definition_namespaces: List[Tuple[FeatureDefinitionWrapper, str]]

    def build_realtime_fv_udf_col(
        self,
        input_df: pyspark.sql.DataFrame,
        fdw: FeatureDefinitionWrapper,
        namespace: str,
    ) -> Tuple[Column, List[Column]]:
        """
        Builds a Spark udf for executing a specific RTFV. This runs an RTFV,
        which outputs a single temporary object (dict/map for python mode, json for
        pandas mode). We then deserialize this to get the feature columns.

        We use this function in two phases in parallel across multiple RTFVs:
        1. Run an RTFV to get the tmp object
        2. Select columns from the tmp object

        To support running these two phases in parallel, we use this method
        to output the column for (1) and the columns for (2), and concat them
        all together.

        :return: select_column (the tmp rtfv output col), output_columns (the
        columns of the tmp rtfv output, which map to the output features of an
        rtfv)
        """
        output_schema = schema_to_spark(fdw.view_schema)
        if namespace is None:
            namespace = fdw.name

        # Pass in only the non-internal fields and udf-internal fields
        # corresponding to this particular rtfv
        udf_args = []
        for input_col in input_df.schema:
            if udf_internal(ComputeMode.SPARK) not in input_col.name or fdw.id in input_col.name:
                udf_args.append(input_col.name)
        udf_arg_idx_map = {}
        for arg_idx in range(len(udf_args)):
            udf_arg_idx_map[udf_args[arg_idx]] = arg_idx
        realtime_pipeline = RealtimeFeaturePipeline(
            name=fdw.name,
            fv_id=fdw.id,
            pipeline=fdw.pipeline,
            transformations=fdw.transformations,
            udf_arg_idx_map=udf_arg_idx_map,
            output_schema=output_schema,
            events_df_timestamp_field=self.events_df_timestamp_field,
        )

        from pyspark.sql.functions import col

        output_columns = []
        rtfv_tmp_output_name = f"_{namespace}_rtfv_output"
        for c in output_schema:
            output_column_name = (
                f"{namespace}{fdw.namespace_separator}{c.name}" if self.use_namespace_feature_prefix else c.name
            )
            output_columns.append(col(f"{rtfv_tmp_output_name}.{c.name}").alias(output_column_name))

        from pyspark.sql.functions import from_json
        from pyspark.sql.functions import pandas_udf
        from pyspark.sql.functions import udf
        from pyspark.sql.types import StringType

        if realtime_pipeline.mode == "python":
            # Python features are output as a single dict / map column, so we
            # map that into individual columns
            _rtfv_udf = udf(realtime_pipeline.py_wrapper, output_schema)
            udf_col = _rtfv_udf(*[f"`{c}`" for c in udf_args]).alias(rtfv_tmp_output_name)
            return udf_col, output_columns
        else:
            assert realtime_pipeline.mode == "pandas"
            # Pandas features are output into a single struct, so we deserialize
            # here + cast into multiple columns.
            # Note: from_json will return null in the case of an unparseable
            # string.
            _rtfv_udf = pandas_udf(realtime_pipeline.pandas_udf_wrapper, StringType())
            deserialized_udf_col = from_json(_rtfv_udf(*[f"`{c}`" for c in udf_args]), output_schema)
            return deserialized_udf_col.alias(rtfv_tmp_output_name), output_columns

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        """
        Executes multiple RTFV transformations on the same input dataframe.

        Note: If the user defines their transformation to produce extra columns
        (besides what's specified in output_schema), they will be ignored. If
        there are missing columns they will fail in this function during
        runtime.
        """
        udf_select_columns = []
        rtfv_output_columns = []
        input_df = self.input_node.to_dataframe(spark)
        for fdw, namespace in self.feature_definition_namespaces:
            select_col, output_cols = self.build_realtime_fv_udf_col(input_df, fdw, namespace)
            udf_select_columns.append(select_col)
            rtfv_output_columns.extend(output_cols)

        # Execute rtfvs in parallel, then deserialize outputs into columns
        input_columns = [f"`{c.name}`" for c in input_df.schema]
        rtfv_tmp_outputs = input_df.select(*input_columns, *udf_select_columns)
        return rtfv_tmp_outputs.select(*input_columns, *rtfv_output_columns)


@attrs.frozen
class PipelineEvalSparkNode(SparkExecNode):
    inputs_map: Dict[str, SparkExecNode]
    feature_definition_wrapper: FeatureDefinitionWrapper

    # Needed for correct behavior by tecton_sliding_window udf if it exists in the pipeline
    feature_time_limits: Optional[pendulum.Period]

    check_view_schema: bool

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        spark_pipeline = SparkFeaturePipeline(
            spark,
            self.feature_definition_wrapper.pipeline,
            transformations=self.feature_definition_wrapper.transformations,
            materialization_context_limits=self.feature_time_limits,
            schedule_interval=self.feature_definition_wrapper.batch_materialization_schedule,
            data_source_inputs={k: self.inputs_map[k].to_dataframe(spark) for k in self.inputs_map},
            output_schema=schema_to_spark(self.feature_definition_wrapper.view_schema),
        )
        df = spark_pipeline.get_dataframe()
        if self.feature_time_limits is None and self.feature_definition_wrapper.materialization_start_timestamp:
            df = df.filter(
                df[self.feature_definition_wrapper.timestamp_key]
                >= self.feature_definition_wrapper.materialization_start_timestamp
            )

        if self.check_view_schema:
            validate_df_columns_and_feature_types(
                df, self.feature_definition_wrapper.view_schema, allow_extraneous_columns=False
            )

        return df


@attrs.frozen
class PartialAggSparkNode(SparkExecNode):
    input_node: SparkExecNode
    fdw: FeatureDefinitionWrapper = attrs.field()
    window_start_column_name: str
    aggregation_tile_interval: pendulum.Duration
    window_end_column_name: Optional[str] = None
    aggregation_anchor_time: Optional[datetime] = None

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = partial_aggregations.construct_partial_time_aggregation_df(
            self.input_node.to_dataframe(spark),
            self.fdw.partial_aggregate_group_by_columns,
            self.fdw.trailing_time_window_aggregation(aggregation_slide_period=self.aggregation_tile_interval),
            self.fdw.get_feature_store_format_version,
            self.fdw.view_schema,
            window_start_column_name=self.window_start_column_name,
            window_end_column_name=self.window_end_column_name,
            aggregation_anchor_time=self.aggregation_anchor_time,
        )
        return df


@attrs.frozen
class OnlinePartialAggSparkNodeV2(SparkExecNode):
    input_node: SparkExecNode
    fdw: FeatureDefinitionWrapper = attrs.field()
    aggregation_groups: Tuple[AggregationGroup, ...]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = partial_aggregations.construct_online_partial_agg_v2_df(
            self.input_node.to_dataframe(spark),
            [
                *self.fdw.partial_aggregate_group_by_columns,
                query_consts.aggregation_tile_id(),
                query_consts.aggregation_group_id(),
            ],
            self.aggregation_groups,
            time_key=self.fdw.trailing_time_window_aggregation().time_key,
            view_schema=self.fdw.view_schema,
            is_continuous=self.fdw.is_continuous,
        )
        return df


@attrs.frozen
class OnlineListAggSparkNode(SparkExecNode):
    input_node: SparkExecNode
    fdw: FeatureDefinitionWrapper = attrs.field()
    aggregation_groups: Tuple[AggregationGroup, ...]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = partial_aggregations.construct_partial_agg_lists_for_online_df(
            self.input_node.to_dataframe(spark),
            [*self.fdw.partial_aggregate_group_by_columns, query_consts.aggregation_group_id()],
            self.aggregation_groups,
        )
        return df


@attrs.frozen
class MetricsCollectorSparkNode(SparkExecNode):
    input_node: SparkExecNode
    metrics_collector: data_observability.MetricsCollector = attrs.field(
        factory=data_observability.get_active_metrics_collector
    )

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.metrics_collector.observe(self.input_node.to_dataframe(spark))


@attrs.frozen
class StagingSparkNode(SparkExecNode):
    input_node: SparkExecNode
    staging_table_name: str
    query_tree_step: QueryTreeStep

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        # TODO(danny): consider implementing this in Spark, but for now this is unnecessary and is a passthrough
        return self.input_node.to_dataframe(spark)


@attrs.frozen
class PythonDataSparkNode(SparkExecNode):
    columns: Tuple[str, ...]
    data: Tuple[Tuple[Any, ...]]
    schema: Optional[Schema] = None

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        if self.schema:
            spark_schema = schema_to_spark(self.schema)
            return spark.createDataFrame(self.data, spark_schema)
        return spark.createDataFrame(self.data, self.columns)



from typing import Optional

from tecton._internals import metadata_service
from tecton_core import conf
from tecton_core.offline_store import DEFAULT_OPTIONS_PROVIDERS
from tecton_core.offline_store import OfflineStoreOptionsProvider
from tecton_core.offline_store import S3Options
from tecton_core.offline_store import S3OptionsSource
from tecton_proto.common.id__client_pb2 import Id
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetOfflineStoreCredentialsRequest


class MDSOfflineStoreOptionsProvider(OfflineStoreOptionsProvider):
    def get_s3_options_for_feature_view(self, feature_view_id: Id) -> Optional[S3Options]:
        if conf.get_bool("USE_LOCAL_OFFLINE_STORE_CREDENTIALS"):
            return None
        request = GetOfflineStoreCredentialsRequest(feature_view_id=feature_view_id)
        response = metadata_service.instance().GetOfflineStoreCredentials(request)
        if not response.HasField("aws"):
            return None
        return S3Options(
            access_key_id=response.aws.access_key_id,
            secret_access_key=response.aws.secret_access_key,
            session_token=response.aws.session_token,
            region=conf.get_or_raise("CLUSTER_REGION"),
            options_source=S3OptionsSource.MDS,
        )

    def get_s3_options_for_data_source(self, data_source_id: Id) -> Optional[S3Options]:
        if conf.get_bool("USE_LOCAL_OFFLINE_STORE_CREDENTIALS"):
            return None
        request = GetOfflineStoreCredentialsRequest(data_source_id=data_source_id)
        response = metadata_service.instance().GetOfflineStoreCredentials(request)
        if not response.HasField("aws"):
            return None
        return S3Options(
            access_key_id=response.aws.access_key_id,
            secret_access_key=response.aws.secret_access_key,
            session_token=response.aws.session_token,
            region=conf.get_or_raise("CLUSTER_REGION"),
            options_source=S3OptionsSource.MDS,
        )


INTERACTIVE_OFFLINE_STORE_OPTIONS_PROVIDERS = [MDSOfflineStoreOptionsProvider(), *DEFAULT_OPTIONS_PROVIDERS]



import datetime
import logging
import os
from typing import Optional

from pyspark.sql import SparkSession

from tecton_core import feature_view_utils
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.time_utils import align_time_downwards
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_materialization.materialization_utils import get_statsd_client
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadata
from tecton_proto.materialization.job_metadata__client_pb2 import OfflineStoreType
from tecton_proto.materialization.job_metadata__client_pb2 import OnlineStoreType
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


logger = logging.getLogger(__name__)


def get_materialized_feature_columns(fd: FeatureDefinitionWrapper):
    return feature_view_utils.get_input_feature_columns(
        fd.view_schema.to_proto(),
        fd.join_keys,
        fd.timestamp_key,
    )


def export_consumption_metrics(
    spark: SparkSession,
    fd: FeatureDefinitionWrapper,
    row_count,
    params: MaterializationTaskParams,
    store_type,
    job_metadata_client: JobMetadataClient,
    online_store_type: OnlineStoreType = None,
    only_new_consumption: bool = False,
):
    assert store_type in ("offline", "online")
    # we only use this for dynamo now. no easy way to get the online_store_type
    assert store_type != "online" or online_store_type == OnlineStoreType.ONLINE_STORE_TYPE_DYNAMO
    # Don't want to call fd.features, since in the
    # temporal aggregate case it will return a feature for each time window,
    # which is not how the data is actually materialized.
    materialized_feature_columns = get_materialized_feature_columns(fd)

    num_rows = row_count
    num_values = num_rows * len(materialized_feature_columns)

    if not only_new_consumption:
        statsd_client = get_statsd_client(spark)

        statsd_safe_fd_name = fd.name.replace(":", "__")
        statsd_safe_workspace = fd.workspace.replace(":", "__")
        statsd_client.incr(
            f"tecton-{store_type}-store.cm_feature_write_rows.{statsd_safe_workspace}.{statsd_safe_fd_name}.", num_rows
        )
        statsd_client.incr(
            f"tecton-{store_type}-store.cm_feature_write_values.{statsd_safe_workspace}.{statsd_safe_fd_name}",
            num_values,
        )

    def updater(job_metadata: JobMetadata) -> Optional[JobMetadata]:
        new_proto = JobMetadata()
        new_proto.CopyFrom(job_metadata)
        aligned_bucket_start = int(
            align_time_downwards(datetime.datetime.now(), datetime.timedelta(hours=1)).timestamp()
        )
        if store_type == "offline":
            offline = new_proto.materialization_consumption_info.offline_store_consumption
            bucket = offline.consumption_info[aligned_bucket_start]

            if params.offline_store_path.startswith("s3"):
                offline.offline_store_type = OfflineStoreType.OFFLINE_STORE_TYPE_S3
            elif params.offline_store_path.startswith("dbfs"):
                offline.offline_store_type = OfflineStoreType.OFFLINE_STORE_TYPE_DBFS
            elif params.offline_store_path.startswith("gs"):
                offline.offline_store_type = OfflineStoreType.OFFLINE_STORE_TYPE_GCS
            elif os.environ.get("TEST_ONLY_TECTON_DYNAMODB_ENDPOINT_OVERRIDE"):
                offline.offline_store_type = OfflineStoreType.OFFLINE_STORE_TYPE_S3
            else:
                msg = f"Unknown offline store type path: {offline.offline_store_type}"
                raise Exception(msg)
        elif store_type == "online":
            online = new_proto.materialization_consumption_info.online_store_consumption
            bucket = online.consumption_info[aligned_bucket_start]

            online.online_store_type = online_store_type
        else:
            exc_msg = f"Unknown store type: {store_type}"
            raise Exception(exc_msg)

        bucket.rows_written += num_rows
        bucket.features_written += num_values

        return new_proto

    if params.HasField("job_metadata_table") and params.use_new_consumption_metrics:
        job_metadata_client.update(updater)

    logger.info(f"Exported {store_type} consumption metrics")


def export_consumption_debug_metrics(
    spark: SparkSession,
    offline_store_count: int,
    dataframe_count: int,
    unhandled_analysis_exception: bool,
):
    statsd_client = get_statsd_client(spark)

    if unhandled_analysis_exception:
        statsd_client.incr("consumption.unhandled_analysis_exception")
    elif offline_store_count != dataframe_count:
        statsd_client.incr("consumption.diff_rows_count", offline_store_count - dataframe_count)
    else:
        statsd_client.incr("consumption.same_rows")



from typing import Dict

from tecton._internals import metadata_service
from tecton_core.errors import TectonNotFoundError
from tecton_core.secrets import SecretCacheKey
from tecton_core.secrets import SecretResolver
from tecton_proto.common.secret__client_pb2 import SecretReference
from tecton_proto.secrets.secrets_service__client_pb2 import GetSecretValueRequest


# Singleton local secret store for Notebook development
_local_secret_store: Dict[SecretCacheKey, str] = {}


def set_local_secret(scope: str, key: str, value: str) -> None:
    """Set the secret in Local Secret Store singleton instance."""
    _local_secret_store[SecretCacheKey(scope, key)] = value


class LocalDevSecretResolver(SecretResolver):
    """Secret Resolver for local development. Secrets are retrieved from Secret Service or locally specified values.

    Example:
        db_password = LocalDevSecretResolver().resolve(SecretReference(scope="prod", key="db_password"))
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def resolve(self, secret: SecretReference) -> str:
        if secret.is_local:
            return self._fetch_local_secret(secret.scope, secret.key)
        return self._fetch_mds_secret(secret.scope, secret.key)

    @staticmethod
    def _fetch_mds_secret(scope: str, key: str) -> str:
        try:
            request = GetSecretValueRequest(scope=scope, key=key)
            response = metadata_service.instance().GetSecretValue(request)
            return response.value
        except TectonNotFoundError:
            msg = f"Failed to get secret `{key}` in scope `{scope}`. Please check if the secret scope or key exists."
            raise RuntimeError(msg)
        except Exception as e:
            msg = f"Failed to get secret `{key}` in scope `{scope}` with {e}"
            raise RuntimeError(msg)

    @staticmethod
    def _fetch_local_secret(scope: str, key: str) -> str:
        return _local_secret_store[SecretCacheKey(scope, key)]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/feature_view__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.args import basic_info__client_pb2 as tecton__proto_dot_args_dot_basic__info____client__pb2
from tecton_proto.args import data_source__client_pb2 as tecton__proto_dot_args_dot_data__source____client__pb2
from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2
from tecton_proto.args import pipeline__client_pb2 as tecton__proto_dot_args_dot_pipeline____client__pb2
from tecton_proto.common import analytics_options__client_pb2 as tecton__proto_dot_common_dot_analytics__options____client__pb2
from tecton_proto.common import compute_mode__client_pb2 as tecton__proto_dot_common_dot_compute__mode____client__pb2
from tecton_proto.common import data_source_type__client_pb2 as tecton__proto_dot_common_dot_data__source__type____client__pb2
from tecton_proto.common import data_type__client_pb2 as tecton__proto_dot_common_dot_data__type____client__pb2
from tecton_proto.common import framework_version__client_pb2 as tecton__proto_dot_common_dot_framework__version____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.common import spark_schema__client_pb2 as tecton__proto_dot_common_dot_spark__schema____client__pb2
from tecton_proto.common import time_window__client_pb2 as tecton__proto_dot_common_dot_time__window____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,tecton_proto/args/feature_view__client.proto\x12\x11tecton_proto.args\x1a\x1egoogle/protobuf/duration.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a*tecton_proto/args/basic_info__client.proto\x1a+tecton_proto/args/data_source__client.proto\x1a,tecton_proto/args/diff_options__client.proto\x1a(tecton_proto/args/pipeline__client.proto\x1a\x33tecton_proto/common/analytics_options__client.proto\x1a.tecton_proto/common/compute_mode__client.proto\x1a\x32tecton_proto/common/data_source_type__client.proto\x1a+tecton_proto/common/data_type__client.proto\x1a\x33tecton_proto/common/framework_version__client.proto\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a.tecton_proto/common/spark_schema__client.proto\x1a-tecton_proto/common/time_window__client.proto\"\x84\r\n\x0f\x46\x65\x61tureViewArgs\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12N\n\x11\x66\x65\x61ture_view_type\x18\x02 \x01(\x0e\x32\".tecton_proto.args.FeatureViewTypeR\x0f\x66\x65\x61tureViewType\x12\x37\n\x04info\x18\x03 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12\x46\n\x07version\x18\x1d \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionB\x05\x92M\x02\x08\x05R\x07version\x12.\n\x0fprevent_destroy\x18  \x01(\x08\x42\x05\x92M\x02\x08\x01R\x0epreventDestroy\x12P\n\x07options\x18\" \x03(\x0b\x32/.tecton_proto.args.FeatureViewArgs.OptionsEntryB\x05\x92M\x02\x08\x01R\x07options\x12H\n\x0c\x63\x61\x63he_config\x18$ \x01(\x0b\x32\x1e.tecton_proto.args.CacheConfigB\x05\x92M\x02\x08\x01R\x0b\x63\x61\x63heConfig\x12G\n\x08\x65ntities\x18\x04 \x03(\x0b\x32$.tecton_proto.args.EntityKeyOverrideB\x05\x92M\x02\x08\x06R\x08\x65ntities\x12|\n\x1ematerialized_feature_view_args\x18\x1c \x01(\x0b\x32..tecton_proto.args.MaterializedFeatureViewArgsB\x05\x92M\x02\x10\x01H\x00R\x1bmaterializedFeatureViewArgs\x12M\n\rrealtime_args\x18\x19 \x01(\x0b\x32\x1f.tecton_proto.args.RealtimeArgsB\x05\x92M\x02\x10\x01H\x00R\x0crealtimeArgs\x12Z\n\x12\x66\x65\x61ture_table_args\x18\x1a \x01(\x0b\x32#.tecton_proto.args.FeatureTableArgsB\x05\x92M\x02\x10\x01H\x00R\x10\x66\x65\x61tureTableArgs\x12\x34\n\x16\x63ontext_parameter_name\x18% \x01(\tR\x14\x63ontextParameterName\x12\x30\n\x14online_serving_index\x18\x05 \x03(\tR\x12onlineServingIndex\x12,\n\x0eonline_enabled\x18\x0e \x01(\x08\x42\x05\x92M\x02 \x0bR\ronlineEnabled\x12.\n\x0foffline_enabled\x18\x0f \x01(\x08\x42\x05\x92M\x02 \x0bR\x0eofflineEnabled\x12Z\n\x12\x62\x61tch_compute_mode\x18# \x01(\x0e\x32%.tecton_proto.common.BatchComputeModeB\x05\x92M\x02\x08\tR\x10\x62\x61tchComputeMode\x12>\n\x08pipeline\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.args.PipelineB\x05\x92M\x02\x08\x06R\x08pipeline\x12`\n\x13\x64\x61ta_quality_config\x18! \x01(\x0b\x32$.tecton_proto.args.DataQualityConfigB\n\x92M\x02\x08\x01\x92M\x02\x10\x01R\x11\x64\x61taQualityConfig\x12N\n\x12\x66orced_view_schema\x18\x1e \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x10\x66orcedViewSchema\x12^\n\x1a\x66orced_materialized_schema\x18\x1f \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x18\x66orcedMaterializedSchema\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x0b\n\ttype_argsJ\x04\x08\x07\x10\x08J\x04\x08\x08\x10\tJ\x04\x08\t\x10\nJ\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0cJ\x04\x08\x0c\x10\rJ\x04\x08\r\x10\x0eJ\x04\x08\x10\x10\x11J\x04\x08\x11\x10\x12J\x04\x08\x12\x10\x13J\x04\x08\x13\x10\x14J\x04\x08\x14\x10\x15J\x04\x08\x15\x10\x16J\x04\x08\x16\x10\x17J\x04\x08\x17\x10\x18J\x04\x08\x18\x10\x19J\x04\x08\x1b\x10\x1c\"f\n\x11\x45ntityKeyOverride\x12\x34\n\tentity_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x08\x65ntityId\x12\x1b\n\tjoin_keys\x18\x02 \x03(\tR\x08joinKeys\"K\n\x0e\x42\x61\x63kfillConfig\x12\x39\n\x04mode\x18\x01 \x01(\x0e\x32%.tecton_proto.args.BackfillConfigModeR\x04mode\"\xce\x01\n\x0cOutputStream\x12)\n\x10include_features\x18\x01 \x01(\x08R\x0fincludeFeatures\x12\x44\n\x07kinesis\x18\x02 \x01(\x0b\x32(.tecton_proto.args.KinesisDataSourceArgsH\x00R\x07kinesis\x12>\n\x05kafka\x18\x03 \x01(\x0b\x32&.tecton_proto.args.KafkaDataSourceArgsH\x00R\x05kafkaB\r\n\x0bstream_sink\"\xab\x13\n\x1bMaterializedFeatureViewArgs\x12.\n\x0ftimestamp_field\x18\x01 \x01(\tB\x05\x92M\x02 \x10R\x0etimestampField\x12G\n\x0e\x62\x61tch_schedule\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationB\x05\x92M\x02 \x08R\rbatchSchedule\x12O\n\x12\x66\x65\x61ture_start_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x05\x92M\x02\x08\x01R\x10\x66\x65\x61tureStartTime\x12i\n manual_trigger_backfill_end_time\x18\x13 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x05\x92M\x02\x08\x01R\x1cmanualTriggerBackfillEndTime\x12\x85\x01\n\x15max_backfill_interval\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationB6\x92M\"* \n\x1emax_batch_aggregation_interval\x92M\t*\x07\x12\x05\x30.7.0\x92M\x02\x08\x01R\x13maxBackfillInterval\x12\x41\n\x0bserving_ttl\x18\x05 \x01(\x0b\x32\x19.google.protobuf.DurationB\x05\x92M\x02\x08\x06R\nservingTtl\x12\x85\x01\n\x14offline_store_legacy\x18\x06 \x01(\x0b\x32,.tecton_proto.args.OfflineFeatureStoreConfigB%\x92M\x02 \r\x92M\x11*\x0f\n\roffline_store\x92M\t*\x07\x12\x05\x30.8.0R\x12offlineStoreLegacy\x12Q\n\roffline_store\x18\x19 \x01(\x0b\x32%.tecton_proto.args.OfflineStoreConfigB\x05\x92M\x02 \x0cR\x0cofflineStore\x12\x45\n\rbatch_compute\x18\x07 \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\x0c\x62\x61tchCompute\x12G\n\x0estream_compute\x18\x08 \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\rstreamCompute\x12O\n\nmonitoring\x18\t \x01(\x0b\x32#.tecton_proto.args.MonitoringConfigB\n\x92M\x02\x08\x01\x92M\x02\x10\x01R\nmonitoring\x12M\n\x10\x64\x61ta_source_type\x18\n \x01(\x0e\x32#.tecton_proto.common.DataSourceTypeR\x0e\x64\x61taSourceType\x12G\n\x0conline_store\x18\x0b \x01(\x0b\x32$.tecton_proto.args.OnlineStoreConfigR\x0bonlineStore\x12\x33\n\x15incremental_backfills\x18\x0c \x01(\x08R\x14incrementalBackfills\x12L\n\x14\x61ggregation_interval\x18\r \x01(\x0b\x32\x19.google.protobuf.DurationR\x13\x61ggregationInterval\x12\x64\n\x16stream_processing_mode\x18\x0f \x01(\x0e\x32\'.tecton_proto.args.StreamProcessingModeB\x05\x92M\x02 \x06R\x14streamProcessingMode\x12I\n\x0c\x61ggregations\x18\x0e \x03(\x0b\x32%.tecton_proto.args.FeatureAggregationR\x0c\x61ggregations\x12\x44\n\routput_stream\x18\x10 \x01(\x0b\x32\x1f.tecton_proto.args.OutputStreamR\x0coutputStream\x12O\n\rbatch_trigger\x18\x11 \x01(\x0e\x32#.tecton_proto.args.BatchTriggerTypeB\x05\x92M\x02\x08\x01R\x0c\x62\x61tchTrigger\x12?\n\x06schema\x18\x12 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaB\n\x92M\x02\x08\x06\x92M\x02 \x10R\x06schema\x12:\n\x19\x61ggregation_secondary_key\x18\x14 \x01(\tR\x17\x61ggregationSecondaryKey\x12s\n\x1csecondary_key_output_columns\x18\x15 \x03(\x0b\x32+.tecton_proto.args.SecondaryKeyOutputColumnB\x05\x92M\x02\x18\x05R\x19secondaryKeyOutputColumns\x12I\n\x1drun_transformation_validation\x18\x16 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x1brunTransformationValidation\x12K\n\x1etecton_materialization_runtime\x18\x17 \x01(\tB\x05\x92M\x02\x08\x01R\x1ctectonMaterializationRuntime\x12J\n\x13lifetime_start_time\x18\x18 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x11lifetimeStartTime\x12-\n\x12\x63ompaction_enabled\x18\x1a \x01(\x08R\x11\x63ompactionEnabled\x12\x32\n\x15stream_tiling_enabled\x18\x1c \x01(\x08R\x13streamTilingEnabled\x12J\n\x10stream_tile_size\x18\x1d \x01(\x0b\x32\x19.google.protobuf.DurationB\x05\x92M\x02\x18\x05R\x0estreamTileSize\x12\'\n\x0b\x65nvironment\x18\x1b \x01(\tB\x05\x92M\x02\x08\x01R\x0b\x65nvironment\x12\x43\n\nattributes\x18\x1e \x03(\x0b\x32\x1c.tecton_proto.args.AttributeB\x05\x92M\x02 \x10R\nattributes\x12<\n\nembeddings\x18\x1f \x03(\x0b\x32\x1c.tecton_proto.args.EmbeddingR\nembeddings\x12<\n\ninferences\x18  \x03(\x0b\x32\x1c.tecton_proto.args.InferenceR\ninferences\"e\n\tAttribute\x12\x16\n\x06\x63olumn\x18\x01 \x01(\tR\x06\x63olumn\x12@\n\x0c\x63olumn_dtype\x18\x02 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x0b\x63olumnDtype\"w\n\tInference\x12@\n\rinput_columns\x18\x01 \x03(\x0b\x32\x1b.tecton_proto.common.ColumnR\x0cinputColumns\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x14\n\x05model\x18\x03 \x01(\tR\x05model\"\x8f\x01\n\tEmbedding\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x16\n\x06\x63olumn\x18\x02 \x01(\tR\x06\x63olumn\x12@\n\x0c\x63olumn_dtype\x18\x03 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x0b\x63olumnDtype\x12\x14\n\x05model\x18\x04 \x01(\tR\x05model\"\xcb\x01\n\x0cRealtimeArgs\x12?\n\x06schema\x18\x02 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaB\x05\x92M\x02 \x10R\x06schema\x12)\n\x0c\x65nvironments\x18\x04 \x03(\tB\x05\x92M\x02\x08\x01R\x0c\x65nvironments\x12\x43\n\nattributes\x18\x05 \x03(\x0b\x32\x1c.tecton_proto.args.AttributeB\x05\x92M\x02 \x10R\nattributesJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04\"\xac\x06\n\x10\x46\x65\x61tureTableArgs\x12?\n\x06schema\x18\x06 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaB\x05\x92M\x02 \x10R\x06schema\x12\x41\n\x0bserving_ttl\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationB\x05\x92M\x02\x08\x06R\nservingTtl\x12\x85\x01\n\x14offline_store_legacy\x18\x07 \x01(\x0b\x32,.tecton_proto.args.OfflineFeatureStoreConfigB%\x92M\x02 \r\x92M\x11*\x0f\n\roffline_store\x92M\t*\x07\x12\x05\x30.8.0R\x12offlineStoreLegacy\x12Q\n\roffline_store\x18\x0c \x01(\x0b\x32%.tecton_proto.args.OfflineStoreConfigB\x05\x92M\x02 \x0cR\x0cofflineStore\x12G\n\x0conline_store\x18\x08 \x01(\x0b\x32$.tecton_proto.args.OnlineStoreConfigR\x0bonlineStore\x12\x45\n\rbatch_compute\x18\t \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\x0c\x62\x61tchCompute\x12O\n\nmonitoring\x18\n \x01(\x0b\x32#.tecton_proto.args.MonitoringConfigB\n\x92M\x02\x08\x01\x92M\x02\x10\x01R\nmonitoring\x12K\n\x1etecton_materialization_runtime\x18\x0b \x01(\tB\x05\x92M\x02\x08\x01R\x1ctectonMaterializationRuntime\x12\x43\n\nattributes\x18\r \x03(\x0b\x32\x1c.tecton_proto.args.AttributeB\x05\x92M\x02 \x10R\nattributes\x12.\n\x0ftimestamp_field\x18\x0e \x01(\tB\x05\x92M\x02 \x10R\x0etimestampFieldJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06\"\xbf\x06\n\x12\x46\x65\x61tureAggregation\x12\x16\n\x06\x63olumn\x18\x01 \x01(\tR\x06\x63olumn\x12\x1a\n\x08\x66unction\x18\x02 \x01(\tR\x08\x66unction\x12\x62\n\x0f\x66unction_params\x18\x05 \x03(\x0b\x32\x39.tecton_proto.args.FeatureAggregation.FunctionParamsEntryR\x0e\x66unctionParams\x12l\n\x12time_window_legacy\x18\x06 \x01(\x0b\x32\x19.google.protobuf.DurationB#\x92M\x0f*\r\n\x0btime_window\x92M\t*\x07\x12\x05\x30.8.0\x92M\x02 \tR\x10timeWindowLegacy\x12\x19\n\x04name\x18\x07 \x01(\tB\x05\x92M\x02 \x01R\x04name\x12G\n\x0btime_window\x18\x08 \x01(\x0b\x32\x1d.tecton_proto.args.TimeWindowB\x05\x92M\x02 \nH\x00R\ntimeWindow\x12N\n\x0flifetime_window\x18\t \x01(\x0b\x32#.tecton_proto.common.LifetimeWindowH\x00R\x0elifetimeWindow\x12S\n\x12time_window_series\x18\n \x01(\x0b\x32#.tecton_proto.args.TimeWindowSeriesH\x00R\x10timeWindowSeries\x12G\n\x0c\x63olumn_dtype\x18\x0b \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeB\x05\x92M\x02 \x10R\x0b\x63olumnDtype\x12Y\n\x18\x62\x61tch_sawtooth_tile_size\x18\x0c \x01(\x0b\x32\x19.google.protobuf.DurationB\x05\x92M\x02\x18\x05R\x15\x62\x61tchSawtoothTileSize\x1a`\n\x13\x46unctionParamsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32\x1d.tecton_proto.args.ParamValueR\x05value:\x02\x38\x01\x42\x08\n\x06windowJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05\"]\n\nParamValue\x12!\n\x0bint64_value\x18\x01 \x01(\x03H\x00R\nint64Value\x12#\n\x0c\x64ouble_value\x18\x02 \x01(\x01H\x00R\x0b\x64oubleValueB\x07\n\x05value\"\x8f\x01\n\x11\x44\x61taQualityConfig\x12\x37\n\x14\x64\x61ta_quality_enabled\x18\x01 \x01(\x08\x42\x05\x92M\x02 \x03R\x12\x64\x61taQualityEnabled\x12\x41\n\x19skip_default_expectations\x18\x02 \x01(\x08\x42\x05\x92M\x02 \x03R\x17skipDefaultExpectations\"\xac\x05\n\rClusterConfig\x12\\\n\x10\x65xisting_cluster\x18\x01 \x01(\x0b\x32(.tecton_proto.args.ExistingClusterConfigB\x05\x92M\x02\x08\x01H\x00R\x0f\x65xistingCluster\x12S\n\x0enew_databricks\x18\x02 \x01(\x0b\x32#.tecton_proto.args.NewClusterConfigB\x05\x92M\x02\x08\x01H\x00R\rnewDatabricks\x12\x45\n\x07new_emr\x18\x03 \x01(\x0b\x32#.tecton_proto.args.NewClusterConfigB\x05\x92M\x02\x08\x01H\x00R\x06newEmr\x12^\n\x0fimplicit_config\x18\x04 \x01(\x0b\x32\'.tecton_proto.args.DefaultClusterConfigB\n\x92M\x02\x08\x01\x92M\x02\x18\x05H\x00R\x0eimplicitConfig\x12V\n\x0fjson_databricks\x18\x05 \x01(\x0b\x32$.tecton_proto.args.JsonClusterConfigB\x05\x92M\x02\x08\x01H\x00R\x0ejsonDatabricks\x12H\n\x08json_emr\x18\x06 \x01(\x0b\x32$.tecton_proto.args.JsonClusterConfigB\x05\x92M\x02\x08\x01H\x00R\x07jsonEmr\x12R\n\rjson_dataproc\x18\x07 \x01(\x0b\x32$.tecton_proto.args.JsonClusterConfigB\x05\x92M\x02\x08\x01H\x00R\x0cjsonDataproc\x12\x41\n\x04rift\x18\x08 \x01(\x0b\x32$.tecton_proto.args.RiftClusterConfigB\x05\x92M\x02\x08\x01H\x00R\x04riftB\x08\n\x06\x63onfig\"@\n\x11JsonClusterConfig\x12+\n\x04json\x18\x01 \x01(\x0b\x32\x17.google.protobuf.StructR\x04json\"G\n\x15\x45xistingClusterConfig\x12.\n\x13\x65xisting_cluster_id\x18\x01 \x01(\tR\x11\x65xistingClusterId\"\x9f\x03\n\x10NewClusterConfig\x12#\n\rinstance_type\x18\x01 \x01(\tR\x0cinstanceType\x12\x33\n\x15instance_availability\x18\x06 \x01(\tR\x14instanceAvailability\x12*\n\x11number_of_workers\x18\x02 \x01(\x05R\x0fnumberOfWorkers\x12\x32\n\x16root_volume_size_in_gb\x18\x03 \x01(\x05R\x12rootVolumeSizeInGb\x12\x34\n\x16\x65xtra_pip_dependencies\x18\x04 \x03(\tR\x14\x65xtraPipDependencies\x12\x41\n\x0cspark_config\x18\x05 \x01(\x0b\x32\x1e.tecton_proto.args.SparkConfigR\x0bsparkConfig\x12&\n\x0f\x66irst_on_demand\x18\x07 \x01(\x05R\rfirstOnDemand\x12\x30\n\x14pinned_spark_version\x18\x08 \x01(\tR\x12pinnedSparkVersion\"8\n\x11RiftClusterConfig\x12#\n\rinstance_type\x18\x01 \x01(\tR\x0cinstanceType\"\xbd\x01\n\x14\x44\x65\x66\x61ultClusterConfig\x12\x38\n\x18\x64\x61tabricks_spark_version\x18\x01 \x01(\tR\x16\x64\x61tabricksSparkVersion\x12*\n\x11\x65mr_spark_version\x18\x02 \x01(\tR\x0f\x65mrSparkVersion\x12?\n\x1ctecton_compute_instance_type\x18\x03 \x01(\tR\x19tectonComputeInstanceType\"\x83\x03\n\x0bSparkConfig\x12.\n\x13spark_driver_memory\x18\x01 \x01(\tR\x11sparkDriverMemory\x12\x32\n\x15spark_executor_memory\x18\x02 \x01(\tR\x13sparkExecutorMemory\x12?\n\x1cspark_driver_memory_overhead\x18\x03 \x01(\tR\x19sparkDriverMemoryOverhead\x12\x43\n\x1espark_executor_memory_overhead\x18\x04 \x01(\tR\x1bsparkExecutorMemoryOverhead\x12L\n\nspark_conf\x18\x05 \x03(\x0b\x32-.tecton_proto.args.SparkConfig.SparkConfEntryR\tsparkConf\x1a<\n\x0eSparkConfEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xe6\x01\n\x11OnlineStoreConfig\x12@\n\x06\x64ynamo\x18\x01 \x01(\x0b\x32&.tecton_proto.args.DynamoDbOnlineStoreH\x00R\x06\x64ynamo\x12;\n\x05redis\x18\x02 \x01(\x0b\x32#.tecton_proto.args.RedisOnlineStoreH\x00R\x05redis\x12\x44\n\x08\x62igtable\x18\x03 \x01(\x0b\x32&.tecton_proto.args.BigtableOnlineStoreH\x00R\x08\x62igtableB\x0c\n\nstore_type\"G\n\x13\x44ynamoDbOnlineStore\x12\x18\n\x07\x65nabled\x18\x03 \x01(\x08R\x07\x65nabledJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06\"\xc8\x01\n\x10RedisOnlineStore\x12\x30\n\x10primary_endpoint\x18\x02 \x01(\tB\x05\x92M\x02\x08\x06R\x0fprimaryEndpoint\x12=\n\x14\x61uthentication_token\x18\x04 \x01(\tB\n\x92M\x02\x08\x01\x92M\x02\x18\x06R\x13\x61uthenticationToken\x12\x1f\n\x07\x65nabled\x18\x05 \x01(\x08\x42\x05\x92M\x02\x18\x05R\x07\x65nabledJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\tJ\x04\x08\t\x10\n\"o\n\x13\x42igtableOnlineStore\x12\x18\n\x07\x65nabled\x18\x01 \x01(\x08R\x07\x65nabled\x12\x1d\n\nproject_id\x18\x02 \x01(\tR\tprojectId\x12\x1f\n\x0binstance_id\x18\x03 \x01(\tR\ninstanceId\"\xdb\x01\n\x19OfflineFeatureStoreConfig\x12<\n\x07parquet\x18\x01 \x01(\x0b\x32 .tecton_proto.args.ParquetConfigH\x00R\x07parquet\x12\x36\n\x05\x64\x65lta\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.args.DeltaConfigH\x00R\x05\x64\x65lta\x12:\n\x15subdirectory_override\x18\x03 \x01(\tB\x05\x92M\x02\x08\x01R\x14subdirectoryOverrideB\x0c\n\nstore_type\"\x0f\n\rParquetConfig\"X\n\x0b\x44\x65ltaConfig\x12I\n\x13time_partition_size\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11timePartitionSize\"\x80\x02\n\x12OfflineStoreConfig\x12^\n\x14staging_table_format\x18\x01 \x01(\x0b\x32,.tecton_proto.args.OfflineFeatureStoreConfigR\x12stagingTableFormat\x12\x39\n\x15publish_full_features\x18\x02 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x13publishFullFeatures\x12O\n\x12publish_start_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x05\x92M\x02\x08\x01R\x10publishStartTime\"\xb8\x01\n\x10MonitoringConfig\x12+\n\x11monitor_freshness\x18\x01 \x01(\x08R\x10monitorFreshness\x12H\n\x12\x65xpected_freshness\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11\x65xpectedFreshness\x12\'\n\x0b\x61lert_email\x18\x03 \x01(\tB\x06\xf2\xe2\x02\x02\x08\x01R\nalertEmailJ\x04\x08\x02\x10\x03\"\x9f\x02\n\x18SecondaryKeyOutputColumn\x12@\n\x0btime_window\x18\x01 \x01(\x0b\x32\x1d.tecton_proto.args.TimeWindowH\x00R\ntimeWindow\x12N\n\x0flifetime_window\x18\x03 \x01(\x0b\x32#.tecton_proto.common.LifetimeWindowH\x00R\x0elifetimeWindow\x12S\n\x12time_window_series\x18\x04 \x01(\x0b\x32#.tecton_proto.args.TimeWindowSeriesH\x00R\x10timeWindowSeries\x12\x12\n\x04name\x18\x02 \x01(\tR\x04nameB\x08\n\x06window\"\x83\x01\n\nTimeWindow\x12\x42\n\x0fwindow_duration\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0ewindowDuration\x12\x31\n\x06offset\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x06offset\"\x86\x02\n\x10TimeWindowSeries\x12<\n\x0cseries_start\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0bseriesStart\x12\x38\n\nseries_end\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\tseriesEnd\x12\x36\n\tstep_size\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x08stepSize\x12\x42\n\x0fwindow_duration\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0ewindowDuration\"5\n\x0b\x43\x61\x63heConfig\x12&\n\x0fmax_age_seconds\x18\x01 \x01(\x03R\rmaxAgeSeconds*\xae\x01\n\x0f\x46\x65\x61tureViewType\x12!\n\x1d\x46\x45\x41TURE_VIEW_TYPE_UNSPECIFIED\x10\x00\x12\x1e\n\x1a\x46\x45\x41TURE_VIEW_TYPE_REALTIME\x10\x03\x12#\n\x1f\x46\x45\x41TURE_VIEW_TYPE_FEATURE_TABLE\x10\x04\x12\'\n#FEATURE_VIEW_TYPE_FWV5_FEATURE_VIEW\x10\x05\"\x04\x08\x01\x10\x01\"\x04\x08\x02\x10\x02*\xbf\x01\n\x12\x42\x61\x63kfillConfigMode\x12$\n BACKFILL_CONFIG_MODE_UNSPECIFIED\x10\x00\x12?\n;BACKFILL_CONFIG_MODE_SINGLE_BATCH_SCHEDULE_INTERVAL_PER_JOB\x10\x01\x12\x42\n>BACKFILL_CONFIG_MODE_MULTIPLE_BATCH_SCHEDULE_INTERVALS_PER_JOB\x10\x02*\x8f\x01\n\x14StreamProcessingMode\x12&\n\"STREAM_PROCESSING_MODE_UNSPECIFIED\x10\x00\x12(\n$STREAM_PROCESSING_MODE_TIME_INTERVAL\x10\x01\x12%\n!STREAM_PROCESSING_MODE_CONTINUOUS\x10\x02*\xa8\x01\n\x10\x42\x61tchTriggerType\x12\"\n\x1e\x42\x41TCH_TRIGGER_TYPE_UNSPECIFIED\x10\x00\x12 \n\x1c\x42\x41TCH_TRIGGER_TYPE_SCHEDULED\x10\x01\x12\x1d\n\x19\x42\x41TCH_TRIGGER_TYPE_MANUAL\x10\x02\x12/\n+BATCH_TRIGGER_TYPE_NO_BATCH_MATERIALIZATION\x10\x03\x42\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.feature_view__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _FEATUREVIEWARGS_OPTIONSENTRY._options = None
  _FEATUREVIEWARGS_OPTIONSENTRY._serialized_options = b'8\001'
  _FEATUREVIEWARGS.fields_by_name['info']._options = None
  _FEATUREVIEWARGS.fields_by_name['info']._serialized_options = b'\222M\002\020\001'
  _FEATUREVIEWARGS.fields_by_name['version']._options = None
  _FEATUREVIEWARGS.fields_by_name['version']._serialized_options = b'\222M\002\010\005'
  _FEATUREVIEWARGS.fields_by_name['prevent_destroy']._options = None
  _FEATUREVIEWARGS.fields_by_name['prevent_destroy']._serialized_options = b'\222M\002\010\001'
  _FEATUREVIEWARGS.fields_by_name['options']._options = None
  _FEATUREVIEWARGS.fields_by_name['options']._serialized_options = b'\222M\002\010\001'
  _FEATUREVIEWARGS.fields_by_name['cache_config']._options = None
  _FEATUREVIEWARGS.fields_by_name['cache_config']._serialized_options = b'\222M\002\010\001'
  _FEATUREVIEWARGS.fields_by_name['entities']._options = None
  _FEATUREVIEWARGS.fields_by_name['entities']._serialized_options = b'\222M\002\010\006'
  _FEATUREVIEWARGS.fields_by_name['materialized_feature_view_args']._options = None
  _FEATUREVIEWARGS.fields_by_name['materialized_feature_view_args']._serialized_options = b'\222M\002\020\001'
  _FEATUREVIEWARGS.fields_by_name['realtime_args']._options = None
  _FEATUREVIEWARGS.fields_by_name['realtime_args']._serialized_options = b'\222M\002\020\001'
  _FEATUREVIEWARGS.fields_by_name['feature_table_args']._options = None
  _FEATUREVIEWARGS.fields_by_name['feature_table_args']._serialized_options = b'\222M\002\020\001'
  _FEATUREVIEWARGS.fields_by_name['online_enabled']._options = None
  _FEATUREVIEWARGS.fields_by_name['online_enabled']._serialized_options = b'\222M\002 \013'
  _FEATUREVIEWARGS.fields_by_name['offline_enabled']._options = None
  _FEATUREVIEWARGS.fields_by_name['offline_enabled']._serialized_options = b'\222M\002 \013'
  _FEATUREVIEWARGS.fields_by_name['batch_compute_mode']._options = None
  _FEATUREVIEWARGS.fields_by_name['batch_compute_mode']._serialized_options = b'\222M\002\010\t'
  _FEATUREVIEWARGS.fields_by_name['pipeline']._options = None
  _FEATUREVIEWARGS.fields_by_name['pipeline']._serialized_options = b'\222M\002\010\006'
  _FEATUREVIEWARGS.fields_by_name['data_quality_config']._options = None
  _FEATUREVIEWARGS.fields_by_name['data_quality_config']._serialized_options = b'\222M\002\010\001\222M\002\020\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['timestamp_field']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['timestamp_field']._serialized_options = b'\222M\002 \020'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['batch_schedule']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['batch_schedule']._serialized_options = b'\222M\002 \010'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['feature_start_time']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['feature_start_time']._serialized_options = b'\222M\002\010\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['manual_trigger_backfill_end_time']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['manual_trigger_backfill_end_time']._serialized_options = b'\222M\002\010\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['max_backfill_interval']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['max_backfill_interval']._serialized_options = b'\222M\"* \n\036max_batch_aggregation_interval\222M\t*\007\022\0050.7.0\222M\002\010\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['serving_ttl']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['serving_ttl']._serialized_options = b'\222M\002\010\006'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['offline_store_legacy']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['offline_store_legacy']._serialized_options = b'\222M\002 \r\222M\021*\017\n\roffline_store\222M\t*\007\022\0050.8.0'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['offline_store']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['offline_store']._serialized_options = b'\222M\002 \014'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['monitoring']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['monitoring']._serialized_options = b'\222M\002\010\001\222M\002\020\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['stream_processing_mode']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['stream_processing_mode']._serialized_options = b'\222M\002 \006'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['batch_trigger']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['batch_trigger']._serialized_options = b'\222M\002\010\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['schema']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['schema']._serialized_options = b'\222M\002\010\006\222M\002 \020'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['secondary_key_output_columns']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['secondary_key_output_columns']._serialized_options = b'\222M\002\030\005'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['run_transformation_validation']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['run_transformation_validation']._serialized_options = b'\222M\002\010\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['tecton_materialization_runtime']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['tecton_materialization_runtime']._serialized_options = b'\222M\002\010\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['stream_tile_size']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['stream_tile_size']._serialized_options = b'\222M\002\030\005'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['environment']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['environment']._serialized_options = b'\222M\002\010\001'
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['attributes']._options = None
  _MATERIALIZEDFEATUREVIEWARGS.fields_by_name['attributes']._serialized_options = b'\222M\002 \020'
  _REALTIMEARGS.fields_by_name['schema']._options = None
  _REALTIMEARGS.fields_by_name['schema']._serialized_options = b'\222M\002 \020'
  _REALTIMEARGS.fields_by_name['environments']._options = None
  _REALTIMEARGS.fields_by_name['environments']._serialized_options = b'\222M\002\010\001'
  _REALTIMEARGS.fields_by_name['attributes']._options = None
  _REALTIMEARGS.fields_by_name['attributes']._serialized_options = b'\222M\002 \020'
  _FEATURETABLEARGS.fields_by_name['schema']._options = None
  _FEATURETABLEARGS.fields_by_name['schema']._serialized_options = b'\222M\002 \020'
  _FEATURETABLEARGS.fields_by_name['serving_ttl']._options = None
  _FEATURETABLEARGS.fields_by_name['serving_ttl']._serialized_options = b'\222M\002\010\006'
  _FEATURETABLEARGS.fields_by_name['offline_store_legacy']._options = None
  _FEATURETABLEARGS.fields_by_name['offline_store_legacy']._serialized_options = b'\222M\002 \r\222M\021*\017\n\roffline_store\222M\t*\007\022\0050.8.0'
  _FEATURETABLEARGS.fields_by_name['offline_store']._options = None
  _FEATURETABLEARGS.fields_by_name['offline_store']._serialized_options = b'\222M\002 \014'
  _FEATURETABLEARGS.fields_by_name['monitoring']._options = None
  _FEATURETABLEARGS.fields_by_name['monitoring']._serialized_options = b'\222M\002\010\001\222M\002\020\001'
  _FEATURETABLEARGS.fields_by_name['tecton_materialization_runtime']._options = None
  _FEATURETABLEARGS.fields_by_name['tecton_materialization_runtime']._serialized_options = b'\222M\002\010\001'
  _FEATURETABLEARGS.fields_by_name['attributes']._options = None
  _FEATURETABLEARGS.fields_by_name['attributes']._serialized_options = b'\222M\002 \020'
  _FEATURETABLEARGS.fields_by_name['timestamp_field']._options = None
  _FEATURETABLEARGS.fields_by_name['timestamp_field']._serialized_options = b'\222M\002 \020'
  _FEATUREAGGREGATION_FUNCTIONPARAMSENTRY._options = None
  _FEATUREAGGREGATION_FUNCTIONPARAMSENTRY._serialized_options = b'8\001'
  _FEATUREAGGREGATION.fields_by_name['time_window_legacy']._options = None
  _FEATUREAGGREGATION.fields_by_name['time_window_legacy']._serialized_options = b'\222M\017*\r\n\013time_window\222M\t*\007\022\0050.8.0\222M\002 \t'
  _FEATUREAGGREGATION.fields_by_name['name']._options = None
  _FEATUREAGGREGATION.fields_by_name['name']._serialized_options = b'\222M\002 \001'
  _FEATUREAGGREGATION.fields_by_name['time_window']._options = None
  _FEATUREAGGREGATION.fields_by_name['time_window']._serialized_options = b'\222M\002 \n'
  _FEATUREAGGREGATION.fields_by_name['column_dtype']._options = None
  _FEATUREAGGREGATION.fields_by_name['column_dtype']._serialized_options = b'\222M\002 \020'
  _FEATUREAGGREGATION.fields_by_name['batch_sawtooth_tile_size']._options = None
  _FEATUREAGGREGATION.fields_by_name['batch_sawtooth_tile_size']._serialized_options = b'\222M\002\030\005'
  _DATAQUALITYCONFIG.fields_by_name['data_quality_enabled']._options = None
  _DATAQUALITYCONFIG.fields_by_name['data_quality_enabled']._serialized_options = b'\222M\002 \003'
  _DATAQUALITYCONFIG.fields_by_name['skip_default_expectations']._options = None
  _DATAQUALITYCONFIG.fields_by_name['skip_default_expectations']._serialized_options = b'\222M\002 \003'
  _CLUSTERCONFIG.fields_by_name['existing_cluster']._options = None
  _CLUSTERCONFIG.fields_by_name['existing_cluster']._serialized_options = b'\222M\002\010\001'
  _CLUSTERCONFIG.fields_by_name['new_databricks']._options = None
  _CLUSTERCONFIG.fields_by_name['new_databricks']._serialized_options = b'\222M\002\010\001'
  _CLUSTERCONFIG.fields_by_name['new_emr']._options = None
  _CLUSTERCONFIG.fields_by_name['new_emr']._serialized_options = b'\222M\002\010\001'
  _CLUSTERCONFIG.fields_by_name['implicit_config']._options = None
  _CLUSTERCONFIG.fields_by_name['implicit_config']._serialized_options = b'\222M\002\010\001\222M\002\030\005'
  _CLUSTERCONFIG.fields_by_name['json_databricks']._options = None
  _CLUSTERCONFIG.fields_by_name['json_databricks']._serialized_options = b'\222M\002\010\001'
  _CLUSTERCONFIG.fields_by_name['json_emr']._options = None
  _CLUSTERCONFIG.fields_by_name['json_emr']._serialized_options = b'\222M\002\010\001'
  _CLUSTERCONFIG.fields_by_name['json_dataproc']._options = None
  _CLUSTERCONFIG.fields_by_name['json_dataproc']._serialized_options = b'\222M\002\010\001'
  _CLUSTERCONFIG.fields_by_name['rift']._options = None
  _CLUSTERCONFIG.fields_by_name['rift']._serialized_options = b'\222M\002\010\001'
  _SPARKCONFIG_SPARKCONFENTRY._options = None
  _SPARKCONFIG_SPARKCONFENTRY._serialized_options = b'8\001'
  _REDISONLINESTORE.fields_by_name['primary_endpoint']._options = None
  _REDISONLINESTORE.fields_by_name['primary_endpoint']._serialized_options = b'\222M\002\010\006'
  _REDISONLINESTORE.fields_by_name['authentication_token']._options = None
  _REDISONLINESTORE.fields_by_name['authentication_token']._serialized_options = b'\222M\002\010\001\222M\002\030\006'
  _REDISONLINESTORE.fields_by_name['enabled']._options = None
  _REDISONLINESTORE.fields_by_name['enabled']._serialized_options = b'\222M\002\030\005'
  _OFFLINEFEATURESTORECONFIG.fields_by_name['subdirectory_override']._options = None
  _OFFLINEFEATURESTORECONFIG.fields_by_name['subdirectory_override']._serialized_options = b'\222M\002\010\001'
  _OFFLINESTORECONFIG.fields_by_name['publish_full_features']._options = None
  _OFFLINESTORECONFIG.fields_by_name['publish_full_features']._serialized_options = b'\222M\002\010\001'
  _OFFLINESTORECONFIG.fields_by_name['publish_start_time']._options = None
  _OFFLINESTORECONFIG.fields_by_name['publish_start_time']._serialized_options = b'\222M\002\010\001'
  _MONITORINGCONFIG.fields_by_name['alert_email']._options = None
  _MONITORINGCONFIG.fields_by_name['alert_email']._serialized_options = b'\362\342\002\002\010\001'
  _FEATUREVIEWTYPE._serialized_start=11796
  _FEATUREVIEWTYPE._serialized_end=11970
  _BACKFILLCONFIGMODE._serialized_start=11973
  _BACKFILLCONFIGMODE._serialized_end=12164
  _STREAMPROCESSINGMODE._serialized_start=12167
  _STREAMPROCESSINGMODE._serialized_end=12310
  _BATCHTRIGGERTYPE._serialized_start=12313
  _BATCHTRIGGERTYPE._serialized_end=12481
  _FEATUREVIEWARGS._serialized_start=766
  _FEATUREVIEWARGS._serialized_end=2434
  _FEATUREVIEWARGS_OPTIONSENTRY._serialized_start=2261
  _FEATUREVIEWARGS_OPTIONSENTRY._serialized_end=2319
  _ENTITYKEYOVERRIDE._serialized_start=2436
  _ENTITYKEYOVERRIDE._serialized_end=2538
  _BACKFILLCONFIG._serialized_start=2540
  _BACKFILLCONFIG._serialized_end=2615
  _OUTPUTSTREAM._serialized_start=2618
  _OUTPUTSTREAM._serialized_end=2824
  _MATERIALIZEDFEATUREVIEWARGS._serialized_start=2827
  _MATERIALIZEDFEATUREVIEWARGS._serialized_end=5302
  _ATTRIBUTE._serialized_start=5304
  _ATTRIBUTE._serialized_end=5405
  _INFERENCE._serialized_start=5407
  _INFERENCE._serialized_end=5526
  _EMBEDDING._serialized_start=5529
  _EMBEDDING._serialized_end=5672
  _REALTIMEARGS._serialized_start=5675
  _REALTIMEARGS._serialized_end=5878
  _FEATURETABLEARGS._serialized_start=5881
  _FEATURETABLEARGS._serialized_end=6693
  _FEATUREAGGREGATION._serialized_start=6696
  _FEATUREAGGREGATION._serialized_end=7527
  _FEATUREAGGREGATION_FUNCTIONPARAMSENTRY._serialized_start=7409
  _FEATUREAGGREGATION_FUNCTIONPARAMSENTRY._serialized_end=7505
  _PARAMVALUE._serialized_start=7529
  _PARAMVALUE._serialized_end=7622
  _DATAQUALITYCONFIG._serialized_start=7625
  _DATAQUALITYCONFIG._serialized_end=7768
  _CLUSTERCONFIG._serialized_start=7771
  _CLUSTERCONFIG._serialized_end=8455
  _JSONCLUSTERCONFIG._serialized_start=8457
  _JSONCLUSTERCONFIG._serialized_end=8521
  _EXISTINGCLUSTERCONFIG._serialized_start=8523
  _EXISTINGCLUSTERCONFIG._serialized_end=8594
  _NEWCLUSTERCONFIG._serialized_start=8597
  _NEWCLUSTERCONFIG._serialized_end=9012
  _RIFTCLUSTERCONFIG._serialized_start=9014
  _RIFTCLUSTERCONFIG._serialized_end=9070
  _DEFAULTCLUSTERCONFIG._serialized_start=9073
  _DEFAULTCLUSTERCONFIG._serialized_end=9262
  _SPARKCONFIG._serialized_start=9265
  _SPARKCONFIG._serialized_end=9652
  _SPARKCONFIG_SPARKCONFENTRY._serialized_start=9592
  _SPARKCONFIG_SPARKCONFENTRY._serialized_end=9652
  _ONLINESTORECONFIG._serialized_start=9655
  _ONLINESTORECONFIG._serialized_end=9885
  _DYNAMODBONLINESTORE._serialized_start=9887
  _DYNAMODBONLINESTORE._serialized_end=9958
  _REDISONLINESTORE._serialized_start=9961
  _REDISONLINESTORE._serialized_end=10161
  _BIGTABLEONLINESTORE._serialized_start=10163
  _BIGTABLEONLINESTORE._serialized_end=10274
  _OFFLINEFEATURESTORECONFIG._serialized_start=10277
  _OFFLINEFEATURESTORECONFIG._serialized_end=10496
  _PARQUETCONFIG._serialized_start=10498
  _PARQUETCONFIG._serialized_end=10513
  _DELTACONFIG._serialized_start=10515
  _DELTACONFIG._serialized_end=10603
  _OFFLINESTORECONFIG._serialized_start=10606
  _OFFLINESTORECONFIG._serialized_end=10862
  _MONITORINGCONFIG._serialized_start=10865
  _MONITORINGCONFIG._serialized_end=11049
  _SECONDARYKEYOUTPUTCOLUMN._serialized_start=11052
  _SECONDARYKEYOUTPUTCOLUMN._serialized_end=11339
  _TIMEWINDOW._serialized_start=11342
  _TIMEWINDOW._serialized_end=11473
  _TIMEWINDOWSERIES._serialized_start=11476
  _TIMEWINDOWSERIES._serialized_end=11738
  _CACHECONFIG._serialized_start=11740
  _CACHECONFIG._serialized_end=11793
# @@protoc_insertion_point(module_scope)



import functools
import os

import requests


@functools.lru_cache(maxsize=None)
def session() -> requests.Session:
    """A requests Session to use for all HTTP Requests from the SDK.

    This session is pre-configured with global settings such as client certificates where required.
    Tecton-backend-specific stuff should _not_ be added here, because we also use this session for requests to e.g.
    AWS and Okta.
    """
    s = requests.Session()
    # Allow settings HTTP client certificates from environment variables. This is useful for making requests through
    # the Teleport proxy.
    if "TECTON_CLIENT_CERT" in os.environ:
        s.cert = (os.environ["TECTON_CLIENT_CERT"], os.environ["TECTON_CLIENT_KEY"])
    return s



from unittest import TestCase

from tecton._internals.spark_utils import EXECUTOR_ENV_VARIABLES
from tecton_core import conf


class SparkUtilsTest(TestCase):
    def test_config_presence(self):
        config_vars = set(conf._VALID_KEYS_TO_ALLOWED_SOURCES)
        # Make sure all executor environment variables are defined as valid config keys
        for e in EXECUTOR_ENV_VARIABLES:
            assert e in config_vars



from typing import Dict
from typing import Optional

import attrs
import pendulum
from typeguard import typechecked

from tecton_core import id_helper
from tecton_core.specs import utils
from tecton_proto.args import basic_info__client_pb2 as basic_info_pb2
from tecton_proto.common import framework_version__client_pb2 as framework_version_pb2
from tecton_proto.common import id__client_pb2 as id_pb2
from tecton_proto.data import fco_metadata__client_pb2 as fco_metadata_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


__all__ = [
    "TectonObjectSpec",
]


@utils.frozen_strict
class TectonObjectMetadataSpec:
    """Spec for all basic Tecton object metadata that should be common to all Tecton objects."""

    name: str
    id: str

    description: Optional[str]
    tags: Dict[str, str]
    owner: Optional[str]
    created_at: Optional[pendulum.DateTime] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    workspace: Optional[str] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    defined_in: Optional[str] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    framework_version: framework_version_pb2.FrameworkVersion.ValueType

    # True if this spec represents an object that was defined locally, as opposed to an "applied" object definition
    # retrieved from the backend.
    is_local_object: bool = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    @property
    def id_proto(self) -> id_pb2.Id:
        return id_helper.IdHelper.from_string(self.id)

    @classmethod
    @typechecked
    def from_data_proto(cls, id: id_pb2.Id, fco_metadata: fco_metadata_pb2.FcoMetadata) -> "TectonObjectMetadataSpec":
        return cls(
            name=utils.get_field_or_none(fco_metadata, "name"),
            id=id_helper.IdHelper.to_string(id),
            description=utils.get_field_or_none(fco_metadata, "description"),
            tags=dict(fco_metadata.tags),
            owner=utils.get_field_or_none(fco_metadata, "owner"),
            created_at=utils.get_timestamp_field_or_none(fco_metadata, "created_at"),
            workspace=utils.get_field_or_none(fco_metadata, "workspace"),
            defined_in=utils.get_field_or_none(fco_metadata, "source_filename"),
            framework_version=utils.get_field_or_none(fco_metadata, "framework_version"),
            is_local_object=False,
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, id: id_pb2.Id, basic_info: basic_info_pb2.BasicInfo) -> "TectonObjectMetadataSpec":
        return cls(
            name=utils.get_field_or_none(basic_info, "name"),
            id=id_helper.IdHelper.to_string(id),
            description=utils.get_field_or_none(basic_info, "description"),
            tags=dict(basic_info.tags),
            owner=utils.get_field_or_none(basic_info, "owner"),
            created_at=None,
            workspace=None,
            defined_in=None,
            framework_version=framework_version_pb2.FrameworkVersion.FWV5,
            is_local_object=True,
        )


@utils.frozen_strict
class TectonObjectSpec:
    """Base class for all Tecton object (aka First Class Objects or FCO) specs.

    Specs provide a unified, frozen (i.e. immutable), and more useful abstraction over args and data protos for use
    within the Python SDK.

    See the RFC;
    https://www.notion.so/tecton/RFC-Unified-SDK-for-Notebook-Driven-Development-a377af9d320f46488ea238e51e2ce656
    """

    metadata: TectonObjectMetadataSpec

    # validation_args are used during local FCO validation.
    # They contain the FCO args and some derived data such as derived schemas.
    # validation_args is only set on the TectonObjectSpec for specs created from data protos, i.e. remote specs.
    validation_args: Optional[validator_pb2.FcoValidationArgs] = attrs.field(
        metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True}
    )

    @property
    def name(self) -> str:
        """Convenience accessor."""
        return self.metadata.name

    @property
    def id(self) -> str:
        """Convenience accessor."""
        return self.metadata.id

    @property
    def id_proto(self) -> id_pb2.Id:
        """Convenience accessor."""
        return self.metadata.id_proto

    @property
    def workspace(self) -> Optional[str]:
        """Convenience accessor."""
        return self.metadata.workspace

    @property
    def is_local_object(self) -> bool:
        """Convenience accessor."""
        return self.metadata.is_local_object



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/databricks_api/secrets__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1tecton_proto/databricks_api/secrets__client.proto\x12\x1btecton_proto.databricks_api\"*\n\x12ScopeCreateRequest\x12\x14\n\x05scope\x18\x01 \x01(\tR\x05scope\"`\n\x18ListSecretScopesResponse\x12\x44\n\x06scopes\x18\x01 \x03(\x0b\x32,.tecton_proto.databricks_api.SecretScopeInfoR\x06scopes\"H\n\x0fSecretScopeInfo\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12!\n\x0c\x62\x61\x63kend_type\x18\x02 \x01(\tR\x0b\x62\x61\x63kendType\"]\n\x10SecretPutRequest\x12\x14\n\x05scope\x18\x01 \x01(\tR\x05scope\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\x12!\n\x0cstring_value\x18\x03 \x01(\tR\x0bstringValue\"i\n\x13SecretAclPutRequest\x12\x14\n\x05scope\x18\x01 \x01(\tR\x05scope\x12\x1c\n\tprincipal\x18\x02 \x01(\tR\tprincipal\x12\x1e\n\npermission\x18\x03 \x01(\tR\npermission\"Z\n\nPutRequest\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\x12\x1a\n\x08\x63ontents\x18\x02 \x01(\x0cR\x08\x63ontents\x12\x1c\n\toverwrite\x18\x03 \x01(\x08R\toverwriteB\x1d\n\x19\x63om.tecton.databricks_apiP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.databricks_api.secrets__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.tecton.databricks_apiP\001'
  _SCOPECREATEREQUEST._serialized_start=82
  _SCOPECREATEREQUEST._serialized_end=124
  _LISTSECRETSCOPESRESPONSE._serialized_start=126
  _LISTSECRETSCOPESRESPONSE._serialized_end=222
  _SECRETSCOPEINFO._serialized_start=224
  _SECRETSCOPEINFO._serialized_end=296
  _SECRETPUTREQUEST._serialized_start=298
  _SECRETPUTREQUEST._serialized_end=391
  _SECRETACLPUTREQUEST._serialized_start=393
  _SECRETACLPUTREQUEST._serialized_end=498
  _PUTREQUEST._serialized_start=500
  _PUTREQUEST._serialized_end=590
# @@protoc_insertion_point(module_scope)



from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import Optional


class AuthProvider(ABC):
    """Base library class that provide authentication."""

    @abstractmethod
    def get_auth_header(self) -> Optional[str]:
        msg = "AuthProvider.get_auth_header() is not implemented"
        raise NotImplementedError(msg)


class RequestProvider(ABC):
    """Base library class that provide request headers and url."""

    def __init__(self, auth_provider: AuthProvider) -> None:
        self.auth_provider = auth_provider

    @abstractmethod
    def request_headers(self) -> Dict[str, str]:
        msg = "RequestProvider.request_headers() is not implemented"
        raise NotImplementedError(msg)

    @abstractmethod
    def request_url(self) -> str:
        msg = "RequestProvider.request_url() is not implemented"
        raise NotImplementedError(msg)



import json
import sys
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Set
from typing import Tuple

import click

from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton.cli import printer
from tecton.cli.command import TectonGroup
from tecton_core.errors import TectonAPIValidationError
from tecton_proto.auth.authorization_service__client_pb2 import Assignment
from tecton_proto.auth.authorization_service__client_pb2 import AssignRolesRequest
from tecton_proto.auth.authorization_service__client_pb2 import GetRolesRequest
from tecton_proto.auth.authorization_service__client_pb2 import ListAssignedPrincipalsRequest
from tecton_proto.auth.authorization_service__client_pb2 import ListAssignedRolesRequest
from tecton_proto.auth.authorization_service__client_pb2 import ListAssignedRolesResponse
from tecton_proto.auth.authorization_service__client_pb2 import UnassignRolesRequest
from tecton_proto.auth.principal__client_pb2 import PrincipalType
from tecton_proto.auth.resource__client_pb2 import ResourceType
from tecton_proto.auth.resource_role_assignments__client_pb2 import RoleAssignmentType
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetUserRequest


RESOURCE_TYPES = {
    "workspace": ResourceType.RESOURCE_TYPE_WORKSPACE,
    "organization": ResourceType.RESOURCE_TYPE_ORGANIZATION,
    "secret-scope": ResourceType.RESOURCE_TYPE_SECRET_SCOPE,
}


def _get_role_definitions():
    request = GetRolesRequest()
    response = metadata_service.instance().GetRoles(request)
    return response.roles


@click.command("access-control", cls=TectonGroup)
def access_control():
    """Manage Access Controls."""


@access_control.command("assign-role")
@click.option(
    "-w", "--workspace", required=False, help="Assign role to a specific workspace (default is current workspace)"
)
@click.option("--all-workspaces", required=False, is_flag=True, help="Assign role to all workspaces")
@click.option("-c", "--secret-scope", required=False, help="Assign role to a specific secret scope")
# we can't make the role help dynamic without making top level usage of the CLI make a network request
# since even lazy loading following https://github.com/pallets/click/pull/2348 doesn't work for help text
@click.option(
    "-r", "--role", required=True, type=str, help="Role name (e.g. admin, owner, editor, consumer, viewer, etc)"
)
@click.option("-u", "--user", default=None, help="User Email")
@click.option("-s", "--service-account", default=None, help="Service Account ID")
@click.option("-g", "--principal-group", default=None, help="Principal Group ID")
@click.option("-f", "--file", default=None, help="Newline separated list of user emails.", type=click.File("r"))
def assign_role_command(workspace, all_workspaces, secret_scope, role, user, service_account, principal_group, file):
    """Assign a role to a principal."""
    workspace_resource, workspace_principal = _validate_assign_unassign_input(
        workspace, all_workspaces, secret_scope, user, service_account, principal_group, file
    )

    if file is not None:
        if user or service_account or principal_group:
            msg = "Please use exactly one of: --user, --service-account, --principal-group, or --file"
            raise click.ClickException(msg)
        _bulk_update_user_role(workspace_resource, all_workspaces, secret_scope, file, role)
    else:
        _update_role(
            workspace_resource,
            all_workspaces,
            secret_scope,
            role,
            user,
            service_account,
            principal_group,
            workspace_principal,
        )


@access_control.command()
@click.option(
    "-w", "--workspace", required=False, help="Unassign role to a specific workspace (default is current workspace)"
)
@click.option("--all-workspaces", required=False, is_flag=True, help="Unassign role to all workspaces")
@click.option("-c", "--secret-scope", required=False, help="Assign role to a specific secret scope")
# we can't make the role help dynamic without making top level usage of the CLI make a network request
# since even lazy loading following https://github.com/pallets/click/pull/2348 doesn't work for help text
@click.option(
    "-r", "--role", required=True, type=str, help="Role name (e.g. admin, owner, editor, consumer, viewer, etc)"
)
@click.option("-u", "--user", default=None, help="User Email")
@click.option("-s", "--service-account", default=None, help="Service Account ID")
@click.option("-g", "--principal-group", default=None, help="Principal Group ID")
def unassign_role(workspace, all_workspaces, secret_scope, role, user, service_account, principal_group):
    """Unassign a role from a principal."""
    workspace_resource, workspace_principal = _validate_assign_unassign_input(
        workspace, all_workspaces, secret_scope, user, service_account, principal_group, None
    )

    _update_role(
        workspace_resource,
        all_workspaces,
        secret_scope,
        role,
        user,
        service_account,
        principal_group,
        workspace_principal,
        unassign=True,
    )


def _validate_assign_unassign_input(
    workspace, all_workspaces, secret_scope, user, service_account, principal_group, file
) -> Tuple[Optional[str], Optional[str]]:
    """
    Validate input parameters to assign-role/unassign-role.
    Also mutate 1 parameter workspace into either workspace_resource or workspace_principal.

    :return: pair of workspace_resource, workspace_principal
    """
    if all_workspaces and (workspace or secret_scope):
        msg = "The --all_workspaces flag cannot be used with --workspace or --secret_scope. Please assign roles separately."
        raise click.ClickException(msg)
        # --workspace and --secret_scope can be used together when workspace is the principal
    if workspace and secret_scope:
        workspace_principal = workspace
        workspace_resource = None
        if user or service_account or principal_group or file:
            msg = (
                "When using --secret-scope and --workspace to grant a workspace access to a secret scope, other principal flags "
                "--user, --service-account, --principal-group, --file must be removed."
            )
            raise click.ClickException(msg)
    else:
        workspace_principal = None
        workspace_resource = workspace
    return (workspace_resource, workspace_principal)


def _update_role(
    workspace_resource,
    all_workspaces,
    secret_scope,
    role,
    user,
    service_account,
    principal_group,
    workspace_principal,
    unassign=False,
):
    role = role.lower()
    assignment = Assignment()
    principal_type, principal_id = get_principal_details(user, service_account, principal_group, workspace_principal)

    if role == "admin":
        if all_workspaces or workspace_resource or secret_scope:
            msg = (
                "'Admin' is a cluster-wide role. Please remove --workspace, --all-workspace, and --secret-scope options"
            )
            raise click.ClickException(msg)
        resource_type = ResourceType.RESOURCE_TYPE_ORGANIZATION
    elif all_workspaces:
        resource_type = ResourceType.RESOURCE_TYPE_ORGANIZATION
    elif workspace_resource:
        resource_type = ResourceType.RESOURCE_TYPE_WORKSPACE
        assignment.resource_id = workspace_resource
    elif secret_scope:
        resource_type = ResourceType.RESOURCE_TYPE_SECRET_SCOPE
        assignment.resource_id = secret_scope
    else:
        msg = "Must specify either --workspace, --all-workspaces, or --secret-scope"
        raise click.ClickException(msg)

    role_defs = _get_role_definitions()
    if resource_type == ResourceType.RESOURCE_TYPE_SECRET_SCOPE and not any(
        ResourceType.RESOURCE_TYPE_SECRET_SCOPE in r.assignable_on_resource_types for r in role_defs
    ):
        msg = (
            "Tecton Secrets is not enabled on your cluster. "
            + "Currently, Secrets is only available on the Rift platform. Please contact Tecton Support for more assistance."
        )
        raise click.ClickException(msg)
    role_def = next(
        (r for r in role_defs if r.id == role and _is_role_assignable(r, principal_type, resource_type)), None
    )
    if role_def is None:
        msg = f"Invalid role. Possible values are: {', '.join(r.id for r in role_defs if _is_role_assignable(r, principal_type, resource_type))}"
        raise click.ClickException(msg)

    assignment.resource_type = resource_type
    assignment.principal_type = principal_type
    assignment.principal_id = principal_id
    assignment.role = role_def.legacy_id
    if user is not None:
        human_readable_principal_name = user
    elif service_account is not None:
        human_readable_principal_name = service_account
    elif principal_group is not None:
        human_readable_principal_name = principal_group
    else:
        human_readable_principal_name = workspace_principal
    try:
        if unassign:
            request = UnassignRolesRequest()
            request.assignments.append(assignment)
            metadata_service.instance().UnassignRoles(request)
        else:
            request = AssignRolesRequest()
            request.assignments.append(assignment)
            metadata_service.instance().AssignRoles(request)
        printer.safe_print(f"Successfully updated role for [{human_readable_principal_name}]")
    except Exception as e:
        printer.safe_print(f"Failed to update role for [{human_readable_principal_name}]: {e}", file=sys.stderr)
        sys.exit(1)


def _bulk_update_user_role(workspace, all_workspaces, secret_scope, file, role):
    for user in [line.strip() for line in file.readlines() if len(line.strip()) > 0]:
        _update_role(
            workspace,
            all_workspaces,
            secret_scope,
            role,
            user,
            service_account=None,
            principal_group=None,
            workspace_principal=None,
        )


def _is_role_assignable(role_def, principal_type, resource_type):
    return (
        principal_type in role_def.assignable_to_principal_types
        and resource_type in role_def.assignable_on_resource_types
    )


def get_roles(principal_type, principal_id, resource_type):
    request = ListAssignedRolesRequest()
    request.principal_type = principal_type
    request.principal_id = principal_id
    request.resource_type = resource_type
    response = metadata_service.instance().ListAssignedRoles(request)
    return response


def display_table(headings, roles):
    table = Displayable.from_table(headings=headings, rows=roles, max_width=0)
    # Align columns in the middle horizontally
    table._text_table.set_cols_align(["c" for _ in range(len(headings))])
    printer.safe_print(table)


@dataclass
class ResourceWithRoleAssignments:
    resource_id: Optional[str]

    # sorted list of roles (order will be preserved)
    roles_sorted: List[str]
    # set of roles that have direct assignments
    directly_assigned_roles: Set[str]
    # role to a list of group names that the role is assigned through
    group_assignments_by_role: Mapping[str, List[str]]


@access_control.command("get-roles")
@click.option("-u", "--user", default=None, help="User Email")
@click.option("-s", "--service-account", default=None, help="Service Account ID")
@click.option("-g", "--principal-group", default=None, help="Principal Group ID")
@click.option("-w", "--workspace", default=None, help="Workspace Principal Name")
@click.option(
    "-r",
    "--resource_type",
    default=None,
    type=click.Choice(RESOURCE_TYPES.keys()),
    help="Optional Resource Type to which the Principal has roles assigned.",
)
@click.option("--json-out", default=False, is_flag=True, help="Format Output as JSON")
def get_assigned_roles(user, service_account, principal_group, workspace, resource_type, json_out):
    """Get the roles assigned to a principal."""
    if resource_type is not None:
        resource_type = RESOURCE_TYPES[resource_type]
    principal_type, principal_id = get_principal_details(user, service_account, principal_group, workspace)

    role_defs = _get_role_definitions()

    ws_roles_response = None
    org_roles_response = None
    try:
        if resource_type is None or resource_type == ResourceType.RESOURCE_TYPE_WORKSPACE:
            ws_roles_response = get_roles(principal_type, principal_id, ResourceType.RESOURCE_TYPE_WORKSPACE)
        if resource_type is None or resource_type == ResourceType.RESOURCE_TYPE_ORGANIZATION:
            org_roles_response = get_roles(principal_type, principal_id, ResourceType.RESOURCE_TYPE_ORGANIZATION)
    except Exception as e:
        printer.safe_print(f"Failed to Get Roles: {e}", file=sys.stderr)
        sys.exit(1)

    scope_roles_response = None
    try:
        if resource_type is None or resource_type == ResourceType.RESOURCE_TYPE_SECRET_SCOPE:
            scope_roles_response = get_roles(principal_type, principal_id, ResourceType.RESOURCE_TYPE_SECRET_SCOPE)
    except TectonAPIValidationError as e:
        # TODO: remove when Secrets manager is enabled on every cluster
        # Secrets not enabled on this cluster yet
        scope_roles_response = ListAssignedRolesResponse(assignments=[])
    except Exception as e:
        printer.safe_print(f"Failed to Get Roles: {e}", file=sys.stderr)
        sys.exit(1)

    ws_roles: List[ResourceWithRoleAssignments] = []
    org_roles = ResourceWithRoleAssignments(None, [], set(), {})
    scope_roles: List[ResourceWithRoleAssignments] = []

    ws_roles_response = list(ws_roles_response.assignments) if ws_roles_response else []
    org_roles_response = list(org_roles_response.assignments) if org_roles_response else []
    scope_roles_response = list(scope_roles_response.assignments) if scope_roles_response else []
    for assignment in ws_roles_response + org_roles_response + scope_roles_response:
        roles_for_resource: List[str] = []
        roles_assigned_directly: Set[str] = set()
        role_to_group_names: Mapping[str, List[str]] = {}
        for role_granted in assignment.roles_granted:
            role = _maybe_convert_legacy_role_id(role_defs, role_granted.role)
            groups_with_role = []
            for role_source in role_granted.role_assignment_sources:
                if role_source.assignment_type == RoleAssignmentType.ROLE_ASSIGNMENT_TYPE_DIRECT:
                    roles_assigned_directly.add(role)
                elif role_source.assignment_type == RoleAssignmentType.ROLE_ASSIGNMENT_TYPE_FROM_PRINCIPAL_GROUP:
                    group_name = role_source.principal_group_name
                    groups_with_role.append(group_name)
            roles_for_resource.append(role)
            role_to_group_names[role] = groups_with_role

        if len(roles_for_resource) > 0:
            if assignment.resource_type == ResourceType.RESOURCE_TYPE_WORKSPACE:
                ws_roles.append(
                    ResourceWithRoleAssignments(
                        assignment.resource_id,
                        roles_for_resource,
                        roles_assigned_directly,
                        role_to_group_names,
                    )
                )
            elif assignment.resource_type == ResourceType.RESOURCE_TYPE_ORGANIZATION:
                for role in roles_for_resource:
                    if role not in org_roles.roles_sorted:
                        org_roles.roles_sorted.append(role)
                org_roles.directly_assigned_roles.update(roles_assigned_directly)
                org_roles.group_assignments_by_role.update(role_to_group_names)
            elif assignment.resource_type == ResourceType.RESOURCE_TYPE_SECRET_SCOPE:
                scope_roles.append(
                    ResourceWithRoleAssignments(
                        assignment.resource_id,
                        roles_for_resource,
                        roles_assigned_directly,
                        role_to_group_names,
                    )
                )

    # roles are sorted server side, but re-sort in case org roles came from 2 separate calls to getAssignedRoles
    # (workspace roles also return org roles because of roles on all workspaces)
    org_roles.roles_sorted = sorted(org_roles.roles_sorted)

    if json_out:
        json_output = _convert_assigned_roles_to_json(
            ws_roles=ws_roles,
            org_roles=org_roles,
            scope_roles=scope_roles,
        )
        printer.safe_print(json.dumps(json_output, indent=4))
    else:
        _pretty_print_get_roles_output(
            ws_roles=ws_roles,
            org_roles=org_roles,
            scope_roles=scope_roles,
            show_principal_group_col=workspace is None,
        )


@access_control.command("get-principals")
@click.option("-w", "--workspace", required=False, help="Workspace name")
@click.option("-c", "--secret-scope", required=False, help="Secret scope name")
def get_assigned_principals(workspace, secret_scope):
    """Get principals with access to a resource."""
    if workspace and secret_scope:
        msg = "At most one of --workspace and --secret-scope can be used at the same time."

    request = ListAssignedPrincipalsRequest()
    if workspace:
        request.resource_type = ResourceType.RESOURCE_TYPE_WORKSPACE
        request.resource_id = workspace
    elif secret_scope:
        request.resource_type = ResourceType.RESOURCE_TYPE_SECRET_SCOPE
        request.resource_id = secret_scope
    else:
        request.resource_type = ResourceType.RESOURCE_TYPE_ORGANIZATION
    response = metadata_service.instance().ListAssignedPrincipals(request)

    headings = ["Principal Type", "Principal", "Role", "All Workspaces", "Assigned Directly", "Assigned via Groups"]
    if not workspace:
        headings = [h for h in headings if h != "All Workspaces"]

    output_rows = []
    for assignment_basic_v2 in response.assignments:
        principal = assignment_basic_v2.principal
        if principal.HasField("user"):
            principal_type = "User"
            principal_id_or_email = principal.user.login_email
        elif principal.HasField("service_account"):
            principal_type = "Service Account"
            principal_id_or_email = principal.service_account.id
        elif principal.HasField("group"):
            principal_type = "Group"
            principal_id_or_email = principal.group.id  # customers may complain about the id
        else:
            principal_type = "Workspace"
            principal_id_or_email = principal.workspace.name

        for role_assignment in assignment_basic_v2.role_assignments:
            resource_type = role_assignment.resource_type
            all_workspaces = workspace and resource_type == ResourceType.RESOURCE_TYPE_ORGANIZATION

            for roles_granted in role_assignment.roles_granted:
                role = roles_granted.role
                assigned_directly = False
                assigned_via_groups = []
                for source in roles_granted.role_assignment_sources:
                    if source.assignment_type == RoleAssignmentType.ROLE_ASSIGNMENT_TYPE_DIRECT:
                        assigned_directly = True
                    elif source.assignment_type == RoleAssignmentType.ROLE_ASSIGNMENT_TYPE_FROM_PRINCIPAL_GROUP:
                        assigned_via_groups.append(source.principal_group_name)

                row = [principal_type, principal_id_or_email, role]
                if workspace:
                    row += ["yes" if all_workspaces else ""]
                row += ["directly" if assigned_directly else "", ",".join(assigned_via_groups)]
                output_rows.append(tuple(row))
    display_table(headings, output_rows)


def _convert_assigned_roles_json_generic_resource(
    resource_type_for_output: str,
    # This parameter is is needed for legacy reasons, because for workspaces we called the field "workspace_name"
    # In 1.0 we can remove this parameter and just call the field "name"
    resource_name_field_name: Optional[str],
    roles: List[ResourceWithRoleAssignments],
) -> List[Dict]:
    json_output = []
    for resource in roles:
        roles_granted = []
        for role in resource.roles_sorted:
            assignment_sources = []
            if role in resource.directly_assigned_roles:
                assignment_sources.append({"assignment_type": "DIRECT"})
            for group_name in resource.group_assignments_by_role[role]:
                assignment_sources.append({"assignment_type": "PRINCIPAL_GROUP", "group_name": group_name})
            roles_granted.append({"role": role, "assignment_sources": assignment_sources})
        if len(roles_granted) > 0:
            if resource_name_field_name is None:
                json_output.append(
                    {
                        "resource_type": resource_type_for_output,
                        "roles_granted": roles_granted,
                    }
                )
            else:
                json_output.append(
                    {
                        "resource_type": resource_type_for_output,
                        resource_name_field_name: resource.resource_id,
                        "roles_granted": roles_granted,
                    }
                )
    return json_output


def _convert_assigned_roles_to_json(
    ws_roles: List[ResourceWithRoleAssignments],
    org_roles: ResourceWithRoleAssignments,
    scope_roles: List[ResourceWithRoleAssignments],
) -> List[Dict]:
    json_output = []
    json_output += _convert_assigned_roles_json_generic_resource("WORKSPACE", "workspace_name", ws_roles)
    json_output += _convert_assigned_roles_json_generic_resource("ORGANIZATION", None, [org_roles])
    json_output += _convert_assigned_roles_json_generic_resource("SECRET_SCOPE", "name", scope_roles)
    return json_output


def _pretty_print_assigned_roles_generic_resource(
    resource_type_for_display: str,
    roles: List[ResourceWithRoleAssignments],
    show_principal_groups_col: bool,
):
    """For generic resources of which there can be many (e.g. workspace, secret scope)"""
    headings = [resource_type_for_display, "Role", "Assigned Directly"]
    if show_principal_groups_col:
        headings += ["Assigned via Groups"]
    display_rows = []
    for assignment_for_resource in roles:
        resource_name = assignment_for_resource.resource_id
        already_displayed_resource_name = False
        for role in assignment_for_resource.roles_sorted:
            resource_name_to_display = "" if already_displayed_resource_name else resource_name
            already_displayed_resource_name = True
            assigned_directly = "direct" if role in assignment_for_resource.directly_assigned_roles else ""
            resource_role_row = (resource_name_to_display, role, assigned_directly)
            if show_principal_groups_col:
                group_names = ", ".join(assignment_for_resource.group_assignments_by_role[role])
                resource_role_row += (group_names,)

            display_rows.append(resource_role_row)
    display_table(headings, display_rows)


def _pretty_print_assigned_roles_org_resource(
    org_roles: ResourceWithRoleAssignments,
):
    """For the singleton organization resource"""
    headings = ["Organization Roles", "Assigned Directly", "Assigned via Groups"]
    display_rows = []
    for role in org_roles.roles_sorted:
        assigned_directly = "direct" if role in org_roles.directly_assigned_roles else ""
        group_names = ", ".join(org_roles.group_assignments_by_role[role])
        role_row = (role, assigned_directly, group_names)

        display_rows.append(role_row)
    display_table(headings, display_rows)


def _pretty_print_get_roles_output(
    ws_roles: List[ResourceWithRoleAssignments],
    org_roles: ResourceWithRoleAssignments,
    scope_roles: List[ResourceWithRoleAssignments],
    show_principal_group_col: bool,
):
    if len(ws_roles) > 0:
        _pretty_print_assigned_roles_generic_resource("Workspace", ws_roles, True)
        printer.safe_print()
    if len(org_roles.roles_sorted) > 0:
        _pretty_print_assigned_roles_org_resource(org_roles)
        printer.safe_print()
    if len(scope_roles) > 0:
        _pretty_print_assigned_roles_generic_resource("Secret Scope", scope_roles, show_principal_group_col)


def _maybe_convert_legacy_role_id(role_defs, id):
    role_def = next((r for r in role_defs if r.id == id or r.legacy_id == id), None)
    role_id = id if role_def is None else role_def.id
    return role_id


def get_user_id(email):
    try:
        request = GetUserRequest()
        request.email = email
        response = metadata_service.instance().GetUser(request)
        return response.user.okta_id
    except Exception as e:
        printer.safe_print(f"Failed to Get Roles for email [{email}]: {e}", file=sys.stderr)
        sys.exit(1)


def get_principal_details(user, service_account, principal_group, workspace_principal):
    principal_type_count = sum([p is not None for p in [user, service_account, principal_group, workspace_principal]])
    if principal_type_count > 1:
        msg = "Please mention a single Principal Type using one of --user, --service-account, or --principal-group (or --workspace if using --secret-scope)"
        raise click.ClickException(msg)
    if user:
        return PrincipalType.PRINCIPAL_TYPE_USER, get_user_id(user)
    elif service_account:
        return PrincipalType.PRINCIPAL_TYPE_SERVICE_ACCOUNT, service_account
    elif principal_group:
        return PrincipalType.PRINCIPAL_TYPE_GROUP, principal_group
    elif workspace_principal:
        return PrincipalType.PRINCIPAL_TYPE_WORKSPACE, workspace_principal
    else:
        msg = "Please mention a Principal Type using --user, --service-account, or --principal-group (or --workspace for workspace authorization on secret scopes)."
        raise click.ClickException(msg)



# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import itertools
import os
import platform
import re
import sys
import threading
import traceback
import types

from py4j.clientserver import ClientServer

__all__ = []  # type: ignore


def print_exec(stream):
    ei = sys.exc_info()
    traceback.print_exception(ei[0], ei[1], ei[2], None, stream)


class VersionUtils(object):
    """
    Provides utility method to determine Spark versions with given input string.
    """
    @staticmethod
    def majorMinorVersion(sparkVersion):
        """
        Given a Spark version string, return the (major version number, minor version number).
        E.g., for 2.0.1-SNAPSHOT, return (2, 0).

        Examples
        --------
        >>> sparkVersion = "2.4.0"
        >>> VersionUtils.majorMinorVersion(sparkVersion)
        (2, 4)
        >>> sparkVersion = "2.3.0-SNAPSHOT"
        >>> VersionUtils.majorMinorVersion(sparkVersion)
        (2, 3)
        """
        m = re.search(r'^(\d+)\.(\d+)(\..*)?$', sparkVersion)
        if m is not None:
            return (int(m.group(1)), int(m.group(2)))
        else:
            raise ValueError("Spark tried to parse '%s' as a Spark" % sparkVersion +
                             " version string, but it could not find the major and minor" +
                             " version numbers.")


def fail_on_stopiteration(f):
    """
    Wraps the input function to fail on 'StopIteration' by raising a 'RuntimeError'
    prevents silent loss of data when 'f' is used in a for loop in Spark code
    """
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except StopIteration as exc:
            raise RuntimeError(
                "Caught StopIteration thrown from user's code; failing the task",
                exc
            )

    return wrapper


def walk_tb(tb):
    while tb is not None:
        yield tb
        tb = tb.tb_next


def try_simplify_traceback(tb):
    """
    Simplify the traceback. It removes the tracebacks in the current package, and only
    shows the traceback that is related to the thirdparty and user-specified codes.

    Returns
    -------
    TracebackType or None
      Simplified traceback instance. It returns None if it fails to simplify.

    Notes
    -----
    This keeps the tracebacks once it sees they are from a different file even
    though the following tracebacks are from the current package.

    Examples
    --------
    >>> import importlib
    >>> import sys
    >>> import traceback
    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as tmp_dir:
    ...     with open("%s/dummy_module.py" % tmp_dir, "w") as f:
    ...         _ = f.write(
    ...             'def raise_stop_iteration():\\n'
    ...             '    raise StopIteration()\\n\\n'
    ...             'def simple_wrapper(f):\\n'
    ...             '    def wrapper(*a, **k):\\n'
    ...             '        return f(*a, **k)\\n'
    ...             '    return wrapper\\n')
    ...         f.flush()
    ...         spec = importlib.util.spec_from_file_location(
    ...             "dummy_module", "%s/dummy_module.py" % tmp_dir)
    ...         dummy_module = importlib.util.module_from_spec(spec)
    ...         spec.loader.exec_module(dummy_module)
    >>> def skip_doctest_traceback(tb):
    ...     import pyspark
    ...     root = os.path.dirname(pyspark.__file__)
    ...     pairs = zip(walk_tb(tb), traceback.extract_tb(tb))
    ...     for cur_tb, cur_frame in pairs:
    ...         if cur_frame.filename.startswith(root):
    ...             return cur_tb

    Regular exceptions should show the file name of the current package as below.

    >>> exc_info = None
    >>> try:
    ...     fail_on_stopiteration(dummy_module.raise_stop_iteration)()
    ... except Exception as e:
    ...     tb = sys.exc_info()[-1]
    ...     e.__cause__ = None
    ...     exc_info = "".join(
    ...         traceback.format_exception(type(e), e, tb))
    >>> print(exc_info)  # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
    Traceback (most recent call last):
      File ...
        ...
      File "/.../pyspark/util.py", line ...
        ...
    RuntimeError: ...
    >>> "pyspark/util.py" in exc_info
    True

    If the traceback is simplified with this method, it hides the current package file name:

    >>> exc_info = None
    >>> try:
    ...     fail_on_stopiteration(dummy_module.raise_stop_iteration)()
    ... except Exception as e:
    ...     tb = try_simplify_traceback(sys.exc_info()[-1])
    ...     e.__cause__ = None
    ...     exc_info = "".join(
    ...         traceback.format_exception(
    ...             type(e), e, try_simplify_traceback(skip_doctest_traceback(tb))))
    >>> print(exc_info)  # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
    RuntimeError: ...
    >>> "pyspark/util.py" in exc_info
    False

    In the case below, the traceback contains the current package in the middle.
    In this case, it just hides the top occurrence only.

    >>> exc_info = None
    >>> try:
    ...     fail_on_stopiteration(dummy_module.simple_wrapper(
    ...         fail_on_stopiteration(dummy_module.raise_stop_iteration)))()
    ... except Exception as e:
    ...     tb = sys.exc_info()[-1]
    ...     e.__cause__ = None
    ...     exc_info_a = "".join(
    ...         traceback.format_exception(type(e), e, tb))
    ...     exc_info_b = "".join(
    ...         traceback.format_exception(
    ...             type(e), e, try_simplify_traceback(skip_doctest_traceback(tb))))
    >>> exc_info_a.count("pyspark/util.py")
    2
    >>> exc_info_b.count("pyspark/util.py")
    1
    """
    if "pypy" in platform.python_implementation().lower():
        # Traceback modification is not supported with PyPy in PySpark.
        return None
    if sys.version_info[:2] < (3, 7):
        # Traceback creation is not supported Python < 3.7.
        # See https://bugs.python.org/issue30579.
        return None

    import pyspark

    root = os.path.dirname(pyspark.__file__)
    tb_next = None
    new_tb = None
    pairs = zip(walk_tb(tb), traceback.extract_tb(tb))
    last_seen = []

    for cur_tb, cur_frame in pairs:
        if not cur_frame.filename.startswith(root):
            # Filter the stacktrace from the PySpark source itself.
            last_seen = [(cur_tb, cur_frame)]
            break

    for cur_tb, cur_frame in reversed(list(itertools.chain(last_seen, pairs))):
        # Once we have seen the file names outside, don't skip.
        new_tb = types.TracebackType(
            tb_next=tb_next,
            tb_frame=cur_tb.tb_frame,
            tb_lasti=cur_tb.tb_frame.f_lasti,
            tb_lineno=cur_tb.tb_frame.f_lineno)
        tb_next = new_tb
    return new_tb


def _print_missing_jar(lib_name, pkg_name, jar_name, spark_version):
    print("""
________________________________________________________________________________________________

  Spark %(lib_name)s libraries not found in class path. Try one of the following.

  1. Include the %(lib_name)s library and its dependencies with in the
     spark-submit command as

     $ bin/spark-submit --packages org.apache.spark:spark-%(pkg_name)s:%(spark_version)s ...

  2. Download the JAR of the artifact from Maven Central http://search.maven.org/,
     Group Id = org.apache.spark, Artifact Id = spark-%(jar_name)s, Version = %(spark_version)s.
     Then, include the jar in the spark-submit command as

     $ bin/spark-submit --jars <spark-%(jar_name)s.jar> ...

________________________________________________________________________________________________

""" % {
        "lib_name": lib_name,
        "pkg_name": pkg_name,
        "jar_name": jar_name,
        "spark_version": spark_version
    })


def _parse_memory(s):
    """
    Parse a memory string in the format supported by Java (e.g. 1g, 200m) and
    return the value in MiB

    Examples
    --------
    >>> _parse_memory("256m")
    256
    >>> _parse_memory("2g")
    2048
    """
    units = {'g': 1024, 'm': 1, 't': 1 << 20, 'k': 1.0 / 1024}
    if s[-1].lower() not in units:
        raise ValueError("invalid format: " + s)
    return int(float(s[:-1]) * units[s[-1].lower()])


class InheritableThread(threading.Thread):
    """
    Thread that is recommended to be used in PySpark instead of :class:`threading.Thread`
    when the pinned thread mode is enabled. The usage of this class is exactly same as
    :class:`threading.Thread` but correctly inherits the inheritable properties specific
    to JVM thread such as ``InheritableThreadLocal``.

    Also, note that pinned thread mode does not close the connection from Python
    to JVM when the thread is finished in the Python side. With this class, Python
    garbage-collects the Python thread instance and also closes the connection
    which finishes JVM thread correctly.

    When the pinned thread mode is off, this works as :class:`threading.Thread`.

    .. versionadded:: 3.1.0


    Notes
    -----
    This API is experimental.
    """
    def __init__(self, target, *args, **kwargs):
        from pyspark import SparkContext

        sc = SparkContext._active_spark_context

        if isinstance(sc._gateway, ClientServer):
            # Here's when the pinned-thread mode (PYSPARK_PIN_THREAD) is on.
            properties = sc._jsc.sc().getLocalProperties().clone()
            self._sc = sc

            def copy_local_properties(*a, **k):
                sc._jsc.sc().setLocalProperties(properties)
                return target(*a, **k)

            super(InheritableThread, self).__init__(
                target=copy_local_properties, *args, **kwargs)
        else:
            super(InheritableThread, self).__init__(target=target, *args, **kwargs)

    def __del__(self):
        from pyspark import SparkContext

        if isinstance(SparkContext._gateway, ClientServer):
            thread_connection = self._sc._jvm._gateway_client.thread_connection.connection()
            if thread_connection is not None:
                connections = self._sc._jvm._gateway_client.deque

                # Reuse the lock for Py4J in PySpark
                with SparkContext._lock:
                    for i in range(len(connections)):
                        if connections[i] is thread_connection:
                            connections[i].close()
                            del connections[i]
                            break
                    else:
                        # Just in case the connection was not closed but removed from the queue.
                        thread_connection.close()


if __name__ == "__main__":
    import doctest

    if "pypy" not in platform.python_implementation().lower() and sys.version_info[:2] >= (3, 7):
        (failure_count, test_count) = doctest.testmod()
        if failure_count:
            sys.exit(-1)



import ast
import logging
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

from colorama import Fore
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer

from tecton.cli import printer
from tecton.cli.cli_utils import bold
from tecton.framework import base_tecton_object
from tecton_core.id_helper import IdHelper
from tecton_proto.data.state_update__client_pb2 import ValidationMessage


logger = logging.getLogger(__name__)


# Fallback: if extract_code_block failed to parse the file (maybe due to a syntax error in user
# code), just try to unintelligently extract some lines around lineno.
def extract_lines(file_path: Path, lineno: int) -> Tuple[Optional[str], Optional[int]]:
    with open(file_path) as f:
        lines = list(f.read().splitlines())

        result = []
        start_line = lineno
        for line in reversed(lines[:lineno]):
            result.append(line)
            start_line -= 1
            if line.strip() == "":
                break

        result.reverse()
        for line in lines[lineno:]:
            if line.strip() == "":
                break
            else:
                result.append(line)
        return "\n".join(result), start_line


# For Python stack traces, we only get an approximate line number if the call statement spans
# multiple lines. For example, if exception is thrown inside the constructor call below:
#  1   Something(
#  2        x,
#  3        y,
#  4        z
#  5   )
#
# .. stack frame info might say it is line 3.
#
# To produce nicer error messages, we parse python file and extract the entire code block that spans
# the line we got from the stack frame info.
#
# This function returns code block and the starting line of the code block.
def extract_code_block(file_path: Path, lineno: int) -> Tuple[Optional[str], Optional[int]]:
    with open(file_path) as f:
        source = f.read()

    code = None
    try:
        root = ast.parse(source, str(file_path))

        start_line = None
        end_line = None

        for node in ast.iter_child_nodes(root):
            if node.lineno > lineno:
                end_line = node.lineno - 1
                break
            start_line = node.lineno - 1

        code = "\n".join(source.splitlines()[start_line:end_line])
        return code, start_line
    except Exception as e:
        return extract_lines(file_path, lineno)


@dataclass
class PrettyFrameInfo:
    file_path: Path
    lineno: int
    code_block_start_line: int
    code_block: str


def pretty_error(
    file_path: Path,
    user_file_paths: List[Path],
    repo_root: Path,
    exception,
    error_message,
    error_details=None,
):
    # Print out the raw exception in verbose mode. exc_info=True will print the exception currently being handled.
    logger.debug("Raw exception:", exc_info=True)

    pretty_frames: List[PrettyFrameInfo] = []

    # Otherwise, get error info from the call stack by finding the frame that contains
    # user-provided Python code.
    # TODO: handle the case when there is more than one frame of user code in the stack trace.
    fail_frame = None
    user_files_set = {path.resolve() for path in user_file_paths}

    for frame in traceback.extract_tb(exception.__traceback__):
        frame_file_path = Path(frame.filename).resolve()
        if frame_file_path in user_files_set:
            # Extract code block around the error location
            code, start_line = extract_code_block(frame_file_path, frame.lineno)
            if code is not None and start_line is not None:
                pretty_frames.append(
                    PrettyFrameInfo(
                        file_path=frame_file_path,
                        lineno=frame.lineno,
                        code_block=code,
                        code_block_start_line=start_line,
                    )
                )

    # For a syntax error, call stack is useless but there is error location info stored in the
    # exception itself.
    if isinstance(exception, SyntaxError):
        assert exception.filename is not None and exception.lineno is not None, "Missing information from exception"
        file_path = Path(exception.filename)
        line_no = exception.lineno
        code, start_line = extract_code_block(file_path, line_no)

        if code is not None and start_line is not None:
            pretty_frames.append(
                PrettyFrameInfo(
                    file_path=file_path,
                    lineno=line_no,
                    code_block=code,
                    code_block_start_line=start_line,
                )
            )

    # relative path formatting helper
    def relp(x):
        return str(x.relative_to(repo_root))

    if pretty_frames:
        filename = pretty_frames[-1].file_path
        lineno = pretty_frames[-1].lineno
        printer.safe_print(
            Fore.RED
            + f"Error while processing {bold(relp(filename))}, at line {bold(str(lineno))}: {error_message}"
            + Fore.RESET
        )
    else:
        filename = file_path
        printer.safe_print(Fore.RED + f"Error while processing {bold(relp(filename))}: {error_message}" + Fore.RESET)

    for i, pretty_frame in enumerate(reversed(pretty_frames)):
        if i == 0:
            printer.safe_print(
                f"=================== Around this code block in {bold(relp(pretty_frame.file_path))} ==================="
            )
        else:
            printer.safe_print(
                f"=================== Called from here in {bold(relp(pretty_frame.file_path))} ==================="
            )
        tf = TerminalFormatter(bg="dark", linenos=True)
        tf._lineno = pretty_frame.code_block_start_line
        printer.safe_print(highlight(pretty_frame.code_block, PythonLexer(), tf))

    if error_details:
        printer.safe_print("=================== Error: ===============================")
        printer.safe_print(error_details)


def format_validation_location_lite(obj: base_tecton_object.BaseTectonObject, repo_root: str) -> None:
    printer.safe_print(
        " " * 4,
        f"in {obj.__class__.__name__} {obj.name} declared in {bold(obj._source_info.source_filename + ':' + str(obj._source_info.source_lineno))}\n",
    )


def format_validation_location_fancy(obj: base_tecton_object.BaseTectonObject, repo_root: str) -> None:
    filename = obj._source_info.source_filename
    file_path = Path(repo_root) / filename
    lineno = int(obj._source_info.source_lineno)

    code, start_line = extract_code_block(file_path, lineno)
    tf = TerminalFormatter(bg="dark", linenos=True)
    tf._lineno = start_line
    printer.safe_print(
        f"=================== {obj.__class__.__name__} {obj.name} declared in {bold(filename)} ==================="
    )
    printer.safe_print(highlight(code, PythonLexer(), tf))


def format_server_errors(
    messages: List[ValidationMessage], objects: Sequence[base_tecton_object.BaseTectonObject], repo_root: str
):
    obj_by_id = {}
    for fco_obj in objects:
        obj_by_id[fco_obj.id] = fco_obj

    for i, m in enumerate(messages):
        printer.safe_print(Fore.RED + m.message + Fore.RESET)
        for fco_ref in m.fco_refs:
            obj_id = IdHelper.to_string(fco_ref.fco_id)
            obj = obj_by_id[obj_id]

            # Print the first error with fancy code snippet, then use shorter format for others
            if i == 0 and obj._source_info.source_lineno:
                format_validation_location_fancy(obj, repo_root)
            else:
                format_validation_location_lite(obj, repo_root)



import logging
from typing import Any
from typing import Dict
from typing import Optional

from pyspark.sql import SparkSession

from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.spark_utils import get_or_create_spark_session
from tecton._internals.spark_utils import log_spark_env
from tecton_core import conf


logger = logging.getLogger(__name__)


class TectonContext:
    """
    Execute Spark SQL queries; access various utils.
    """

    _current_context_instance: Optional["TectonContext"] = None
    _config: Dict[str, Any] = {}

    def __init__(self, spark: SparkSession):
        log_spark_env(spark)
        self._spark = spark

    @classmethod
    def _set_config(cls, custom_spark_options=None):
        """
        Sets the configs for TectonContext instance.
        To take effect it must be called before any calls to TectonContext.get_instance().

        :param custom_spark_options: If spark session gets created by TectonContext, custom spark options/
        """
        cls._config = {"custom_spark_options": custom_spark_options}

    @classmethod
    @sdk_public_method
    def get_instance(cls) -> "TectonContext":
        """
        Get the singleton instance of TectonContext.
        """
        # If the instance doesn't exist, creates a new TectonContext from
        # an existing Spark context. Alternatively, creates a new Spark context on the fly.
        if cls._current_context_instance is not None:
            return cls._current_context_instance
        else:
            return cls._generate_and_set_new_instance()

    @classmethod
    def set_global_instance(cls, instance: Optional["TectonContext"]):
        """
        Create the singleton instance of TectonContext from the provided spark session.
        """
        cls._current_context_instance = instance

    @classmethod
    def _generate_and_set_new_instance(cls) -> "TectonContext":
        logger.debug("Generating new Spark session")
        spark = get_or_create_spark_session(
            cls._config.get("custom_spark_options"),
        )
        cls._current_context_instance = cls(spark)
        return cls._current_context_instance

    def _get_spark(self) -> SparkSession:
        return self._spark


@sdk_public_method
def set_tecton_spark_session(spark: SparkSession):
    """
    Configure Tecton to use the provided SparkSession instead of its default.
    """
    TectonContext.set_global_instance(TectonContext(spark))


@sdk_public_method
def get_current_workspace() -> Optional[str]:
    """
    Gets the current workspace set by environment variable, Tecton config file, or `tecton workspace select` command.
    When running `tecton plan` or `tecton apply` Tecton uses the workspace returned by this method.
    """
    return conf.get_or_none("TECTON_WORKSPACE")



from dataclasses import dataclass

import pytest

from tecton.aggregation_functions import AggregationFunction
from tecton.aggregation_functions import approx_count_distinct
from tecton.aggregation_functions import approx_percentile
from tecton.aggregation_functions import first
from tecton.aggregation_functions import first_distinct
from tecton.aggregation_functions import last
from tecton.aggregation_functions import last_distinct


@dataclass
class TestAggregationFunctions__TestCase:
    aggregation: AggregationFunction
    expected_resolved_name: str


TEST_AGGREGATION_FUNCTIONS_TEST_CASES = (
    TestAggregationFunctions__TestCase(
        aggregation=AggregationFunction(base_name="count", resolved_name="count", params={}),
        expected_resolved_name="count",
    ),
    TestAggregationFunctions__TestCase(aggregation=last(3), expected_resolved_name="last_3"),
    TestAggregationFunctions__TestCase(aggregation=last(1003), expected_resolved_name="last_1003"),
    TestAggregationFunctions__TestCase(aggregation=last_distinct(3), expected_resolved_name="last_distinct_3"),
    TestAggregationFunctions__TestCase(aggregation=first(3), expected_resolved_name="first_3"),
    TestAggregationFunctions__TestCase(aggregation=first_distinct(3), expected_resolved_name="first_distinct_3"),
    TestAggregationFunctions__TestCase(
        aggregation=approx_percentile(percentile=0.0, precision=100), expected_resolved_name="approx_percentile_p0_0"
    ),
    TestAggregationFunctions__TestCase(
        aggregation=approx_percentile(percentile=0.0, precision=50), expected_resolved_name="approx_percentile_p0_0"
    ),
    TestAggregationFunctions__TestCase(
        aggregation=approx_percentile(percentile=0.5), expected_resolved_name="approx_percentile_p0_5"
    ),
    TestAggregationFunctions__TestCase(
        aggregation=approx_percentile(percentile=0.99, precision=50), expected_resolved_name="approx_percentile_p0_99"
    ),
    TestAggregationFunctions__TestCase(
        aggregation=approx_percentile(percentile=1.0, precision=50), expected_resolved_name="approx_percentile_p1_0"
    ),
    TestAggregationFunctions__TestCase(
        aggregation=approx_count_distinct(), expected_resolved_name="approx_count_distinct"
    ),
    TestAggregationFunctions__TestCase(
        aggregation=approx_count_distinct(precision=12), expected_resolved_name="approx_count_distinct"
    ),
)


@pytest.mark.parametrize("test_case", TEST_AGGREGATION_FUNCTIONS_TEST_CASES)
def test_aggregation_groups(test_case: TestAggregationFunctions__TestCase):
    assert test_case.aggregation.resolved_name == test_case.expected_resolved_name
